import tensorflow as tf

from neupy.core.properties import (
    ProperFractionProperty,
    ScalarVariableProperty,
    NumberProperty,
)
from .base import GradientDescent


__all__ = ('Adam',)


class Adam(GradientDescent):
    """
    Adam algorithm.

    Parameters
    ----------
    beta1 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    beta2 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-5``.

    step : float
        Learning rate, defaults to ``0.001``.

    {GradientDescent.batch_size}

    {BaseOptimizer.regularizer}

    {BaseOptimizer.network}

    {BaseOptimizer.loss}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {Verbose.verbose}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    References
    ----------
    [1] Diederik P. Kingma, Jimmy Lei Ba
        Adam: a Method for Stochastic Optimization.
        https://arxiv.org/pdf/1412.6980.pdf

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.Adam(network)
    >>> optimizer.train(x_train, y_train)
    """
    step = ScalarVariableProperty(default=0.001)
    beta1 = ProperFractionProperty(default=0.9)
    beta2 = ProperFractionProperty(default=0.999)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_train_updates(self):
        optimizer = tf.train.AdamOptimizer(
            beta1=self.beta1,
            beta2=self.beta2,
            epsilon=self.epsilon,
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]



import tensorflow as tf

from neupy import init
from neupy.utils import as_tuple
from neupy.core.properties import IntProperty, ParameterProperty
from .base import BaseLayer


__all__ = ('Embedding',)


class Embedding(BaseLayer):
    """
    Embedding layer accepts indices as an input and returns
    rows from the weight matrix associated with these indices.
    It's useful when inputs are categorical features or for the
    word embedding tasks.

    Parameters
    ----------
    input_size : int
        Layer's input vector dimension. It's, typically, associated with
        number of categories or number of unique words that input vector has.

    output_size : int
        Layer's output vector dimension.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------

    This example converts dataset that has only categorical
    variables into format that suitable for Embedding layer.

    >>> import numpy as np
    >>> from neupy.layers import *
    >>>
    >>> dataset = np.array([
    ...     ['cold', 'high'],
    ...     ['hot',  'low'],
    ...     ['cold', 'low'],
    ...     ['hot',  'low'],
    ... ])
    >>>
    >>> unique_value, dataset_indices = np.unique(
    ...     dataset, return_inverse=True
    ... )
    >>> dataset_indices = dataset_indices.reshape((4, 2))
    >>> dataset_indices
    array([[0, 1],
           [2, 3],
           [0, 3],
           [2, 3]])
    >>>
    >>> n_features = dataset.shape[1]
    >>> n_unique_categories = len(unique_value)
    >>> embedded_size = 1
    >>>
    >>> network = join(
    ...     Input(n_features),
    ...     Embedding(n_unique_categories, embedded_size),
    ...     # Output from the embedding layer is 3D
    ...     # To make output 2D we need to reshape dimensions
    ...     Reshape(),
    ... )
    """
    input_size = IntProperty(minval=1)
    output_size = IntProperty(minval=1)
    weight = ParameterProperty()

    def __init__(self, input_size, output_size,
                 weight=init.HeNormal(), name=None):

        super(Embedding, self).__init__(name=name)

        self.input_size = input_size
        self.output_size = output_size
        self.weight = weight

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        return input_shape.concatenate(self.output_size)

    def create_variables(self, input_shape):
        self.input_shape = input_shape
        self.weight = self.variable(
            value=self.weight, name='weight',
            shape=as_tuple(self.input_size, self.output_size))

    def output(self, input_value, **kwargs):
        input_value = tf.cast(input_value, tf.int32)
        return tf.gather(self.weight, input_value)

    def __repr__(self):
        return self._repr_arguments(
            self.input_size,
            self.output_size,
            name=self.name,
            weight=self.weight,
        )



import random

import numpy as np
from numpy.core.umath_tests import inner1d

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from .base import DiscreteMemory


__all__ = ('DiscreteBAM',)


def bin2sign(matrix):
    return np.where(matrix == 0, -1, 1)


def sign2bin(matrix):
    return np.where(matrix > 0, 1, 0).astype(int)


def hopfield_energy(weight, X, y):
    return -0.5 * inner1d(X.dot(weight), y)


class DiscreteBAM(DiscreteMemory):
    """
    Discrete BAM Network with associations.
    Network associate every input with some target value.

    Parameters
    ----------
    {DiscreteMemory.Parameters}

    Methods
    -------
    train(X, y)
        Train network and update network weights.

    predict_output(X, n_times=None)
        Using input data recover output data. Returns two arguments.
        First is an input data, second is an output data.

    predict(X, n_times=None)
        Alias to the ``predict_output`` method.

    predict_input(y, n_times=None)
        Using output data recover input data. Returns two arguments.
        First is input data, second is output data.

    energy(X, y)
        Calculate Hopfield Energy for the input and output data.

    Notes
    -----
    - Input and output vectors should contain only binary values.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>> zero.reshape((5, 6))
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> zero_hint = np.matrix([[0, 1, 0, 0]])
    >>> one_hint = np.matrix([[1, 0, 0, 0]])
    >>>
    >>> data = np.concatenate([zero, one], axis=0)
    >>> hints = np.concatenate([zero_hint, one_hint], axis=0)
    >>>
    >>> bamnet = algorithms.DiscreteBAM(mode='sync')
    >>> bamnet.train(data, hints)
    >>>
    >>> recovered_zero, recovered_hint = bamnet.predict(half_zero)
    >>> recovered_hint
    matrix([[0, 1, 0, 0]])
    >>>
    >>> zero_hint
    matrix([[0, 1, 0, 0]])
    """
    def apply_async_process(self, X_sign, y_sign, n_times=None):
        if n_times is None:
            n_times = self.n_times

        n_input_features = X_sign.shape[-1]
        n_output_features = y_sign.shape[-1]

        for _ in range(n_times):
            i = random.randrange(n_input_features)
            j = random.randrange(n_output_features)

            X_sign[:, i] = np.sign(y_sign.dot(self.weight[i, :]))
            y_sign[:, j] = np.sign(X_sign.dot(self.weight[:, j]))

        return sign2bin(X_sign), sign2bin(y_sign)

    def predict_input(self, y_bin, n_times=None):
        if self.weight is None:
            raise NotTrained("Network hasn't been trained yet")

        self.discrete_validation(y_bin)

        y_bin = format_data(y_bin, is_feature1d=False)
        y_sign = bin2sign(y_bin)
        X_sign = np.sign(y_sign.dot(self.weight.T))

        if self.mode == 'sync':
            return sign2bin(X_sign), y_bin

        return self.apply_async_process(X_sign, y_sign, n_times)

    def predict_output(self, X_bin, n_times=None):
        if self.weight is None:
            raise NotTrained("Network hasn't been trained yet")

        self.discrete_validation(X_bin)

        X_bin = format_data(X_bin, is_feature1d=False)
        X_sign = bin2sign(X_bin)
        y_sign = np.sign(X_sign.dot(self.weight))

        if self.mode == 'sync':
            return X_bin, sign2bin(y_sign)

        return self.apply_async_process(X_sign, y_sign, n_times)

    def predict(self, X_bin, n_times=None):
        return self.predict_output(X_bin, n_times)

    def train(self, X_bin, y_bin):
        self.discrete_validation(X_bin)
        self.discrete_validation(y_bin)

        X_sign = bin2sign(format_data(X_bin, is_feature1d=False))
        y_sign = bin2sign(format_data(y_bin, is_feature1d=False))

        _, weight_nrows = X_sign.shape
        _, weight_ncols = y_sign.shape
        weight_shape = (weight_nrows, weight_ncols)

        if self.weight is None:
            self.weight = np.zeros(weight_shape)

        if self.weight.shape != weight_shape:
            raise ValueError(
                "Invalid input shapes. Number of input "
                "features must be equal to {} and {} output "
                "features".format(weight_nrows, weight_ncols))

        self.weight += X_sign.T.dot(y_sign)

    def energy(self, X_bin, y_bin):
        self.discrete_validation(X_bin)
        self.discrete_validation(y_bin)

        X_sign, y_sign = bin2sign(X_bin), bin2sign(y_bin)
        X_sign = format_data(X_sign, is_feature1d=False)
        y_sign = format_data(y_sign, is_feature1d=False)
        nrows, n_features = X_sign.shape

        if nrows == 1:
            return hopfield_energy(self.weight, X_sign, y_sign)

        output = np.zeros(nrows)
        for i, rows in enumerate(zip(X_sign, y_sign)):
            output[i] = hopfield_energy(self.weight, *rows)

        return output



import numpy as np
import tensorflow as tf

from neupy.utils import as_tuple
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import TypedListProperty
from .base import BaseLayer


__all__ = ('Reshape', 'Transpose')


class Reshape(BaseLayer):
    """
    Layer reshapes input tensor.

    Parameters
    ----------
    shape : tuple
        New feature shape. If one dimension specified with the ``-1`` value
        that this dimension will be computed from the total size that remains.
        Defaults to ``-1``.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------

    Covert 4D input to 2D

    >>> from neupy.layers import *
    >>> network = Input((2, 5, 5)) >> Reshape()
    (?, 2, 5, 5) -> [... 2 layers ...] -> (?, 50)

    Convert 3D to 4D

    >>> from neupy.layers import *
    >>> network = Input((5, 4)) >> Reshape((5, 2, 2))
    (?, 5, 4) -> [... 2 layers ...] -> (?, 5, 2, 2)
    """
    shape = TypedListProperty()

    def __init__(self, shape=-1, name=None):
        super(Reshape, self).__init__(name=name)
        self.shape = as_tuple(shape)

        if self.shape.count(-1) >= 2:
            raise ValueError("Only single -1 value can be specified")

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        feature_shape = input_shape[1:]
        missing_value = None

        if -1 in self.shape and feature_shape.is_fully_defined():
            known_shape_values = [val for val in self.shape if val != -1]

            n_feature_values = np.prod(feature_shape.dims)
            n_expected_values = np.prod(known_shape_values)

            if n_feature_values % n_expected_values != 0:
                raise ValueError(
                    "Input shape and specified shape are incompatible Shape: "
                    "{}, Input shape: {}".format(self.shape, input_shape))

            missing_value = int(n_feature_values // n_expected_values)

        n_samples = input_shape[0]
        new_feature_shape = [
            missing_value if val == -1 else val for val in self.shape]

        return tf.TensorShape([n_samples] + new_feature_shape)

    def output(self, input, **kwargs):
        """
        Reshape the feature space for the input value.

        Parameters
        ----------
        input : array-like or Tensorfow variable
        """
        input = tf.convert_to_tensor(input, dtype=tf.float32)
        input_shape = tf.shape(input)

        n_samples = input_shape[0]
        expected_shape = self.get_output_shape(input.shape)
        feature_shape = expected_shape[1:]

        if feature_shape.is_fully_defined():
            # For cases when we have -1 in the shape and feature shape
            # can be precomputed from the input we want to be explicit about
            # expected output shape. Because of the unknown batch dimension
            # it won't be possible for tensorflow to derive exact output
            # shape from the -1
            output_shape = as_tuple(n_samples, feature_shape.dims)
        else:
            output_shape = as_tuple(n_samples, self.shape)

        return tf.reshape(input, output_shape)

    def __repr__(self):
        return self._repr_arguments(self.shape, name=self.name)


class Transpose(BaseLayer):
    """
    Layer transposes input tensor. Permutes the dimensions according
    to the ``perm`` parameter.

    Parameters
    ----------
    perm : tuple or list
        A permutation of the dimensions of the input tensor.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((7, 11)) >> Transpose((0, 2, 1))
    (?, 7, 11) -> [... 2 layers ...] -> (?, 11, 7)
    """
    perm = TypedListProperty()

    def __init__(self, perm, name=None):
        super(Transpose, self).__init__(name=name)
        self.perm = perm

    def fail_if_shape_invalid(self, input_shape):
        if input_shape and max(self.perm) >= input_shape.ndims:
            raise LayerConnectionError(
                "Cannot apply transpose operation to the "
                "input. Permutation: {}, Input shape: {}"
                "".format(self.perm, input_shape))

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        self.fail_if_shape_invalid(input_shape)

        if input_shape.ndims is None:
            n_dims_expected = len(self.perm)
            return tf.TensorShape([None] * n_dims_expected)

        input_shape = np.array(input_shape.dims)
        perm = list(self.perm)

        return tf.TensorShape(input_shape[perm])

    def output(self, input, **kwargs):
        # Input value has batch dimension, but perm will never have it
        # specified as (zero index), so we need to add it in order to
        # fix batch dimension in place.
        return tf.transpose(input, list(self.perm))

    def __repr__(self):
        return self._repr_arguments(self.perm, name=self.name)



import os
import random

import tensorflow as tf
import numpy as np


__all__ = ('as_tuple', 'AttributeKeyDict', 'reproducible')


def as_tuple(*values):
    """
    Convert sequence of values in one big tuple.

    Parameters
    ----------
    *values
        Values that needs to be combined in one big tuple.

    Returns
    -------
    tuple
        All input values combined in one tuple

    Examples
    --------
    >>> as_tuple(None, (1, 2, 3), None)
    (None, 1, 2, 3, None)
    >>>
    >>> as_tuple((1, 2, 3), (4, 5, 6))
    (1, 2, 3, 4, 5, 6)
    """
    cleaned_values = []
    for value in values:
        if isinstance(value, (tuple, list)):
            cleaned_values.extend(value)
        else:
            cleaned_values.append(value)
    return tuple(cleaned_values)


def reproducible(seed=0):
    """
    Set up the same seed value for the NumPy and
    python random module to make your code reproducible.

    Parameters
    ----------
    seed : int
        Defaults to ``0``.
    """
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    random.seed(seed)
    tf.set_random_seed(seed)


class AttributeKeyDict(dict):
    """
    Modified built-in Python ``dict`` class. That modification
    helps get and set values like attributes.

    Examples
    --------
    >>> attrdict = AttributeKeyDict()
    >>> attrdict
    {}
    >>> attrdict.test_key = 'test_value'
    >>> attrdict
    {'test_key': 'test_value'}
    >>> attrdict.test_key
    'test_value'
    """
    def __getattr__(self, attrname):
        return self[attrname]

    def __setattr__(self, attrname, value):
        self[attrname] = value

    def __delattr__(self, attrname):
        del self[attrname]

    def __reduce__(self):
        return (self.__class__, (dict(self),))



import numpy as np

from neupy.utils import format_data
from neupy.core.properties import IntProperty
from neupy.algorithms.base import BaseNetwork


__all__ = ('CMAC',)


class CMAC(BaseNetwork):
    """
    Cerebellar Model Articulation Controller (CMAC) Network based on memory.

    Notes
    -----
    - Network always use Mean Absolute Error (MAE).
    - Network works for multi dimensional target values.

    Parameters
    ----------
    quantization : int
        Network transforms every input to discrete value.
        Quantization value controls number of total number of
        categories after quantization, defaults to ``10``.

    associative_unit_size : int
        Number of associative blocks in memory, defaults to ``2``.

    {BaseNetwork.Parameters}

    Attributes
    ----------
    weight : dict
        Network's weight that contains memorized patterns.

    Methods
    -------
    {BaseSkeleton.predict}

    train(X_train, y_train, X_test=None, y_test=None, epochs=100)
        Trains the network to the data X. Network trains until maximum
        number of ``epochs`` was reached.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy.algorithms import CMAC
    >>>
    >>> train_space = np.linspace(0, 2 * np.pi, 100)
    >>> test_space = np.linspace(np.pi, 2 * np.pi, 50)
    >>>
    >>> X_train = np.reshape(train_space, (100, 1))
    >>> X_test = np.reshape(test_space, (50, 1))
    >>>
    >>> y_train = np.sin(X_train)
    >>> y_test = np.sin(X_test)
    >>>
    >>> cmac = CMAC(
    ...     quantization=100,
    ...     associative_unit_size=32,
    ...     step=0.2,
    ... )
    ...
    >>> cmac.train(X_train, y_train, epochs=100)
    >>>
    >>> predicted_test = cmac.predict(X_test)
    >>> cmac.score(y_test, predicted_test)
    0.0023639417543036569
    """
    quantization = IntProperty(default=10, minval=1)
    associative_unit_size = IntProperty(default=2, minval=2)

    def __init__(self, **options):
        self.weight = {}
        super(CMAC, self).__init__(**options)

    def predict(self, X):
        X = format_data(X)

        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        predicted = []

        for input_sample in self.quantize(X):
            coords = get_memory_coords(input_sample)
            predicted.append(get_result_by_coords(coords))

        return np.array(predicted)

    def get_result_by_coords(self, coords):
        return sum(
            self.weight.setdefault(coord, 0) for coord in coords
        ) / self.associative_unit_size

    def get_memory_coords(self, quantized_value):
        assoc_unit_size = self.associative_unit_size

        for i in range(assoc_unit_size):
            point = ((quantized_value + i) / assoc_unit_size).astype(int)
            yield tuple(np.concatenate([point, [i]]))

    def quantize(self, X):
        return (X * self.quantization).astype(int)

    def one_training_update(self, X_train, y_train):
        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        weight = self.weight
        step = self.step

        n_samples = X_train.shape[0]
        quantized_input = self.quantize(X_train)
        errors = 0

        for input_sample, target_sample in zip(quantized_input, y_train):
            coords = list(get_memory_coords(input_sample))
            predicted = get_result_by_coords(coords)

            error = target_sample - predicted
            for coord in coords:
                weight[coord] += step * error

            errors += sum(abs(error))

        return errors / n_samples

    def score(self, X, y):
        predicted = self.predict(X)
        return np.mean(np.abs(predicted - y))

    def train(self, X_train, y_train, X_test=None, y_test=None, epochs=100):
        is_test_data_partialy_missed = (
            (X_test is None and y_test is not None) or
            (X_test is not None and y_test is None)
        )

        if is_test_data_partialy_missed:
            raise ValueError(
                "Input and target test samples are missed. "
                "They must be defined together or none of them.")

        X_train = format_data(X_train)
        y_train = format_data(y_train)

        if X_test is not None:
            X_test = format_data(X_test)
            y_test = format_data(y_test)

        return super(CMAC, self).train(
            X_train, y_train, X_test, y_test, epochs=epochs)



import numpy as np
import tensorflow as tf

from neupy.utils import (
    tensorflow_session, flatten,
    function_name_scope, make_single_vector,
)
from neupy.core.properties import (BoundedProperty, ChoiceProperty,
                                   WithdrawProperty)
from neupy.algorithms import BaseOptimizer
from neupy.algorithms.gd import objectives
from neupy.utils.tf_utils import setup_parameter_updates


__all__ = ('LevenbergMarquardt',)


@function_name_scope
def compute_jacobian(values, parameters):
    """
    Compute Jacobian matrix.

    Parameters
    ----------
    values : Tensorfow variable
        Computed MSE for each sample separately.

    parameters : list of Tensorfow variable
        Neural network parameters (e.g. weights, biases).

    Returns
    -------
    Tensorfow variable
    """
    values_shape = tf.shape(values)
    n_samples = values_shape[0]

    def compute_gradient_per_value(index, result):
        gradients = tf.gradients(values[index], parameters)
        full_gradient = make_single_vector(gradients)
        return index + 1, result.write(index, full_gradient)

    _, jacobian = tf.while_loop(
        lambda index, _: index < n_samples,
        compute_gradient_per_value,
        [
            tf.constant(0, tf.int32),
            tf.TensorArray(tf.float32, size=n_samples),
        ]
    )

    return jacobian.stack()


class LevenbergMarquardt(BaseOptimizer):
    """
    Levenberg-Marquardt algorithm is a variation of the Newton's method.
    It minimizes MSE error. The algorithm approximates Hessian matrix using
    dot product between two jacobian matrices.

    Notes
    -----
    - Method requires all training data during propagation, which means
      it cannot be trained with mini-batches.

    - Network minimizes only Mean Squared Error (MSE) loss function.

    - Efficient for small training datasets, because it
      computes gradient per each sample separately.

    - Efficient for small-sized networks.

    Parameters
    ----------
    {BaseOptimizer.network}

    mu : float
        Control inversion for J.T * J matrix, defaults to ``0.1``.

    mu_update_factor : float
        Factor to decrease the mu if error was reduced after last update,
        otherwise increase mu by the same factor. Defaults to ``1.2``

    error : {{``mse``}}
        Levenberg-Marquardt works only for quadratic functions.
        Defaults to ``mse``.

    {BaseOptimizer.show_epoch}

    {BaseOptimizer.shuffle_data}

    {BaseOptimizer.signals}

    {BaseOptimizer.verbose}

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.LevenbergMarquardt(network)
    >>> optimizer.train(x_train, y_train)

    See Also
    --------
    :network:`BaseOptimizer` : BaseOptimizer algorithm.
    """
    mu = BoundedProperty(default=0.01, minval=0)
    mu_update_factor = BoundedProperty(default=1.2, minval=1)
    loss = ChoiceProperty(default='mse', choices={'mse': objectives.mse})

    step = WithdrawProperty()
    regularizer = WithdrawProperty()

    def init_functions(self):
        self.variables.update(
            mu=tf.Variable(self.mu, name='lev-marq/mu'),
            last_error=tf.Variable(np.nan, name='lev-marq/last-error'),
        )
        super(LevenbergMarquardt, self).init_functions()

    def init_train_updates(self):
        training_outputs = self.network.training_outputs
        last_error = self.variables.last_error
        error_func = self.variables.loss
        mu = self.variables.mu

        new_mu = tf.where(
            tf.less(last_error, error_func),
            mu * self.mu_update_factor,
            mu / self.mu_update_factor,
        )

        err_for_each_sample = flatten((self.target - training_outputs) ** 2)

        variables = self.network.variables
        params = [var for var in variables.values() if var.trainable]
        param_vector = make_single_vector(params)

        J = compute_jacobian(err_for_each_sample, params)
        J_T = tf.transpose(J)
        n_params = J.shape[1]

        parameter_update = tf.matrix_solve(
            tf.matmul(J_T, J) + new_mu * tf.eye(n_params.value),
            tf.matmul(J_T, tf.expand_dims(err_for_each_sample, 1))
        )
        updated_params = param_vector - flatten(parameter_update)

        updates = [(mu, new_mu)]
        parameter_updates = setup_parameter_updates(params, updated_params)
        updates.extend(parameter_updates)

        return updates

    def one_training_update(self, X_train, y_train):
        if self.errors.train:
            last_error = self.errors.train[-1]
            self.variables.last_error.load(last_error, tensorflow_session())

        return super(LevenbergMarquardt, self).one_training_update(
            X_train, y_train)



import numpy as np

from .base import BaseStepAssociative


__all__ = ('Instar',)


class Instar(BaseStepAssociative):
    """
    Instar is a simple unsupervised Neural Network algorithm
    which detects associations.

    Parameters
    ----------
    {BaseAssociative.Parameters}

    Methods
    -------
    {BaseAssociative.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> train_data = np.array([
    ...     [0, 1, 0, 0],
    ...     [1, 1, 0, 0],
    ... ])
    >>> test_cases = np.array([
    ...     [0, 1, 0, 0],
    ...     [0, 0, 0, 0],
    ...     [0, 0, 1, 1],
    ... ])
    >>>
    >>> instnet = algorithms.Instar(
    ...     n_inputs=4,
    ...     n_outputs=1,
    ...     n_unconditioned=1,
    ...     step=1,
    ...     verbose=False,
    ... )
    >>>
    >>> instnet.train(train_data, epochs=2)
    >>> instnet.predict(test_cases)
    array([[1],
           [0],
           [0]])
    """
    def weight_delta(self, input_row, layer_output):
        n_unconditioned = self.n_unconditioned
        weight = self.weight[n_unconditioned:, :]

        return self.step * np.dot(
            input_row[:, n_unconditioned:].T - weight,
            layer_output.T)



from numpy import dot

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from neupy.core.properties import BoundedProperty
from neupy.algorithms.base import BaseSkeleton
from .utils import pdf_between_data


__all__ = ('GRNN',)


class GRNN(BaseSkeleton):
    """
    Generalized Regression Neural Network (GRNN). Network applies
    only to the regression problems.

    Parameters
    ----------
    std : float
        Standard deviation for PDF function.
        If your input features have high values than standard
        deviation should also be high. For instance, if input features
        from range ``[0, 20]`` that standard deviation should be
        also a big value like ``10`` or ``15``. Small values will
        lead to bad prediction.

    {Verbose.verbose}

    Notes
    -----
    - GRNN Network is sensitive for cases when one input feature
      has higher values than the other one. Input data has to be
      normalized before training.

    - Standard deviation has to match the range of the input features
      Check ``std`` parameter description for more information.

    - The bigger training dataset the slower prediction.
      Algorithm is much more efficient for small datasets.

    - Network uses lazy learning which mean that network doesn't
      need iterative training. It just stores parameters
      and use them to make a predictions.

    Methods
    -------
    train(X_train, y_train, copy=True)
        Network just stores all the information about the data and use
        it for the prediction. Parameter ``copy`` copies input data
        before saving it inside the network.

    predict(X)
        Return prediction per each sample in the ``X``.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.model_selection import train_test_split
    >>> from neupy import algorithms
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     preprocessing.minmax_scale(dataset.data),
    ...     preprocessing.minmax_scale(dataset.target.reshape(-1, 1)),
    ...     test_size=0.3,
    ... )
    >>>
    >>> nw = algorithms.GRNN(std=0.1, verbose=False)
    >>> nw.train(x_train, y_train)
    >>>
    >>> y_predicted = nw.predict(x_test)
    >>> mse = np.mean((y_predicted - y_test) ** 2)
    >>> mse
    0.05280970704568171
    """
    std = BoundedProperty(minval=0)

    def __init__(self, std, verbose=False):
        self.std = std
        self.X_train = None
        self.y_train = None
        super(GRNN, self).__init__(verbose=verbose)

    def train(self, X_train, y_train, copy=True):
        """
        Trains network. PNN doesn't actually train, it just stores
        input data and use it for prediction.

        Parameters
        ----------
        X_train : array-like (n_samples, n_features)

        y_train : array-like (n_samples,)
            Target variable should be vector or matrix
            with one feature column.

        copy : bool
            If value equal to ``True`` than input matrices will
            be copied. Defaults to ``True``.

        Raises
        ------
        ValueError
            In case if something is wrong with input data.
        """
        X_train = format_data(X_train, copy=copy)
        y_train = format_data(y_train, copy=copy)

        if y_train.shape[1] != 1:
            raise ValueError("Target value must be one dimensional array")

        self.X_train = X_train
        self.y_train = y_train

        if X_train.shape[0] != y_train.shape[0]:
            raise ValueError("Number of samples in the input and target "
                             "datasets are different")

    def predict(self, X):
        """
        Make a prediction from the input data.

        Parameters
        ----------
        X : array-like (n_samples, n_features)

        Raises
        ------
        ValueError
            In case if something is wrong with input data.

        Returns
        -------
        array-like (n_samples,)
        """
        if self.X_train is None:
            raise NotTrained(
                "Cannot make a prediction. Network hasn't been trained yet")

        X = format_data(X)

        if X.shape[1] != self.X_train.shape[1]:
            raise ValueError(
                "Input data must contain {0} features, got {1}"
                "".format(self.X_train.shape[1], X.shape[1]))

        ratios = pdf_between_data(self.X_train, X, self.std)
        return (dot(self.y_train.T, ratios) / ratios.sum(axis=0)).T



from __future__ import division

from functools import wraps

import tensorflow as tf

from neupy.utils import asfloat


__all__ = ('define_regularizer', 'l1', 'l2', 'maxnorm')


class Regularizer(object):
    def __init__(self, function, *args, **kwargs):
        self.function = function
        self.exclude = kwargs.pop('exclude', ['bias'])

        self.args = args
        self.kwargs = kwargs

    def __call__(self, network):
        cost = asfloat(0)

        for (layer, varname), variable in network.variables.items():
            if varname not in self.exclude and variable.trainable:
                cost += self.function(variable, *self.args, **self.kwargs)

        return cost

    def __repr__(self):
        kwargs_repr = [repr(arg) for arg in self.args]
        kwargs_repr += ["{}={}".format(k, v) for k, v in self.kwargs.items()]
        kwargs_repr += ["exclude={}".format(self.exclude)]
        return "{}({})".format(self.function.__name__, ', '.join(kwargs_repr))


def define_regularizer(function):
    """
    Wraps regularization function and uses it in order to apply
    regularization per each parameter in the network.

    Examples
    --------
    >>> import tensorflow as tf
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> @algorithms.define_regularizer
    ... def l2(weight, decay_rate=0.01):
    ...     return decay_rate * tf.reduce_sum(tf.pow(weight, 2))
    ...
    >>> l2_regularizer = l2(dacay_rate=0.01)
    >>> network = Input(5) >> Relu(10) >> Sigmoid(1)
    >>> reguarization_cost = l2_regularizer(network)
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        return Regularizer(function, *args, **kwargs)
    return wrapper


@define_regularizer
def l1(weight, decay_rate=0.01):
    """
    Applies l1 regularization to the trainable parameters in the network.

    Regularization cost per weight parameter in the layer can be computed
    in the following way (pseudocode).

    .. code-block:: python

        cost = decay_rate * sum(abs(weight))

    Parameters
    ----------
    decay_rate : float
        Controls training penalties during the parameter updates.
        The larger the value the stronger effect regularization
        has during the training. Defaults to ``0.01``.

    exclude : list
        List of parameter names that has to be excluded from the
        regularization. Defaults to ``['bias']``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.l1(decay_rate=0.01)
    ... )

    With included regularization for bias

    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.l1(decay_rate=0.01, exclude=[])
    ... )
    """
    return tf.multiply(decay_rate, tf.reduce_sum(tf.abs(weight)))


@define_regularizer
def l2(weight, decay_rate=0.01):
    """
    Applies l2 regularization to the trainable parameters in the network.

    Regularization cost per weight parameter in the layer can be computed
    in the following way (pseudocode).

    .. code-block:: python

        cost = decay_rate * sum(weight ** 2)

    Parameters
    ----------
    decay_rate : float
        Controls training penalties during the parameter updates.
        The larger the value the stronger effect regularization
        has during the training. Defaults to ``0.01``.

    exclude : list
        List of parameter names that has to be excluded from the
        regularization. Defaults to ``['bias']``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.l2(decay_rate=0.01)
    ... )

    With included regularization for bias

    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.l2(decay_rate=0.01, exclude=[])
    ... )
    """
    return tf.multiply(decay_rate, tf.reduce_sum(tf.pow(weight, 2)))


@define_regularizer
def maxnorm(weight, decay_rate=0.01):
    """
    Applies max-norm regularization to the trainable parameters in the
    network. Also known and l-inf regularization.

    Regularization cost per weight parameter in the layer can be computed
    in the following way (pseudocode).

    .. code-block:: python

        cost = decay_rate * max(abs(weight))

    Parameters
    ----------
    decay_rate : float
        Controls training penalties during the parameter updates.
        The larger the value the stronger effect regularization
        has during the training. Defaults to ``0.01``.

    exclude : list
        List of parameter names that has to be excluded from the
        regularization. Defaults to ``['bias']``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.maxnorm(decay_rate=0.01)
    ... )

    With included regularization for bias

    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.maxnorm(decay_rate=0.01, exclude=[])
    ... )
    """
    return tf.multiply(decay_rate, tf.reduce_max(tf.abs(weight)))



import tensorflow as tf

from .base import GradientDescent


__all__ = ('Adagrad',)


class Adagrad(GradientDescent):
    """
    Adagrad algorithm.

    Parameters
    ----------
    {GradientDescent.Parameters}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.Adagrad(network)
    >>> optimizer.train(x_train, y_train)

    References
    ----------
    [1] John Duchi, Elad Hazan, Yoram Singer,
        Adaptive Subgradient Methods for Online Learning and Stochastic
        Optimization
        http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
    """
    def init_train_updates(self):
        optimizer = tf.train.AdagradOptimizer(
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]



from __future__ import division

import math

import numpy as np
import progressbar

from neupy.utils.misc import as_tuple


__all__ = (
    'apply_batches', 'minibatches',
    'count_minibatches', 'count_samples',
)


def count_samples(inputs):
    if isinstance(inputs, (list, tuple)):
        return count_samples(inputs[0])
    return len(inputs)


def count_minibatches(inputs, batch_size):
    return int(math.ceil(count_samples(inputs) / batch_size))


def apply_slices(inputs, indices):
    if inputs is None:
        return inputs

    if isinstance(inputs, (list, tuple)):
        return [apply_slices(input_, indices) for input_ in inputs]

    return inputs[indices]


def minibatches(inputs, batch_size=None, shuffle=False):
    """
    Iterates batch slices.

    Parameters
    ----------
    inputs : array-like, list

    batch_size : int
        Mini-batch size. Number should be greater than ``0``.

    shuffle : bool
        Defaults to ``True``.

    Yields
    ------
    object
        Batch slices.
    """
    n_samples = count_samples(inputs)
    batch_size = n_samples if batch_size is None else batch_size
    n_batches = count_minibatches(inputs, batch_size)

    if shuffle:
        indices = np.arange(n_samples)
        np.random.shuffle(indices)

        for index in range(n_batches):
            batch_slice = slice(index * batch_size, (index + 1) * batch_size)
            yield apply_slices(inputs, indices[batch_slice])

    elif n_batches != 1:
        for index in range(n_batches):
            batch_slice = slice(index * batch_size, (index + 1) * batch_size)
            yield apply_slices(inputs, batch_slice)

    else:
        yield inputs


def average_batch_errors(errors, n_samples, batch_size):
    """
    Computes average error per sample. Function assumes that error from
    each batch was just an average loss of each individual sample.

    Parameters
    ----------
    errors : list
        List of average errors calculated per batch.

    n_samples : int
        Number of input samples..

    batch_size : int
        Mini-batch size.

    Returns
    -------
    float
        Average error per sample.
    """
    errors = np.atleast_1d(errors)
    batches = np.ones_like(errors) * batch_size

    if len(errors) == 1:
        return errors.item(0)

    if n_samples % batch_size != 0:
        # Last batch can be smaller than the usual one, because we
        # won't have enough samples for the full one.
        batches[-1] = n_samples % batch_size

    return np.dot(errors, batches) / n_samples


def make_progressbar(max_value, show_output):
    widgets = [
        progressbar.Timer(format='Time: %(elapsed)s'),
        ' |',
        progressbar.Percentage(),
        progressbar.Bar(),
        ' ',
        progressbar.ETA(),
    ]

    if show_output:
        widgets.extend([' | ', progressbar.DynamicMessage('loss')])

    return progressbar.ProgressBar(max_value=max_value, widgets=widgets)


def apply_batches(function, inputs, batch_size, show_progressbar=False,
                  show_output=False, average_outputs=False):
    """
    Splits inputs into mini-batches and passes them to the function.
    Function returns list of outputs or average loss in case
    if ``average_outputs=True``.

    Parameters
    ----------
    function : func
        Function that accepts one or more positional inputs.
        Each of them should be an array-like variable that
        have exactly the same number of rows.

    inputs : tuple, list
        The arguments that will be provided to the function specified
        in the ``function`` argument.

    batch_size : int
        Mini-batch size. Defines maximum number of samples that will be
        used as an input to the ``function``.

    show_progressbar : bool
        When ``True`` than progress bar will be shown in the terminal.
        Defaults to ``False``.

    show_output : bool
        Assumes that outputs from the function errors. The ``True`` value
        will show information in the progressbar. Error will be related to
        the last epoch. Defaults to ``False``.

    average_outputs : bool
        Output from each batch will be combined into single average. This
        option assumes that loss per batch was calculated from.
        Defaults to ``False``.

    Returns
    -------
    list
        List of function outputs.
    """
    n_samples = count_samples(inputs)
    batch_size = n_samples if batch_size is None else batch_size

    n_batches = count_minibatches(inputs, batch_size)
    bar = progressbar.NullBar()

    if show_progressbar and n_batches >= 2:
        bar = make_progressbar(n_batches, show_output)
        bar.update(0)  # triggers empty progressbar

    outputs = []
    iterator = minibatches(inputs, batch_size, shuffle=False)

    for i, sliced_inputs in enumerate(iterator):
        output = function(*as_tuple(sliced_inputs))
        outputs.append(output)

        kwargs = dict(loss=output) if show_output else {}
        bar.update(i, **kwargs)

    # Clean progressbar from the screen
    bar.fd.write('\r' + ' ' * bar.term_width + '\r')

    if average_outputs:
        # When loss calculated per batch separately it might be
        # necessary to combine error into single value
        return average_batch_errors(outputs, n_samples, batch_size)

    return outputs



import numpy as np
import tensorflow as tf

from neupy import init
from neupy.utils import asfloat, as_tuple, tf_utils
from neupy.exceptions import LayerConnectionError, WeightInitializationError
from neupy.core.properties import (
    NumberProperty, TypedListProperty,
    ParameterProperty, IntProperty,
)
from .base import BaseLayer


__all__ = (
    'Linear', 'Sigmoid', 'Tanh', 'Softmax',
    'Relu', 'LeakyRelu', 'Elu', 'PRelu',
    'Softplus', 'HardSigmoid',
)


class Linear(BaseLayer):
    """
    Layer with linear activation function. It applies linear transformation
    when the ``n_units`` parameter specified and acts as an identity
    when it's not specified.

    Parameters
    ----------
    n_units : int or None
        Number of units in the layers. It also corresponds to the number of
        output features that will be produced per sample after passing it
        through this layer. The ``None`` value means that layer will not have
        parameters and it will only apply activation function to the input
        without linear transformation output for the specified input value.
        Defaults to ``None``.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`HeNormal() <neupy.init.HeNormal>`.

    bias : 1D array-like, Tensorfow variable, scalar, Initializer or None
        Defines layer's bias. Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`Constant(0) <neupy.init.Constant>`.
        The ``None`` value excludes bias from the calculations and
        do not add it into parameters list.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    activation_function(input)
        Applies activation function to the input.

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    Linear Regression

    >>> from neupy.layers import *
    >>> network = Input(10) >> Linear(5)
    """
    n_units = IntProperty(minval=1, allow_none=True)
    weight = ParameterProperty()
    bias = ParameterProperty(allow_none=True)

    def __init__(self, n_units=None, weight=init.HeNormal(), bias=0,
                 name=None):

        super(Linear, self).__init__(name=name)

        self.n_units = n_units
        self.weight = weight
        self.bias = bias

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if self.n_units is None:
            return input_shape

        if input_shape and input_shape.ndims != 2:
            raise LayerConnectionError(
                "Input shape expected to have 2 dimensions, got {} instead. "
                "Shape: {}".format(input_shape.ndims, input_shape))

        n_samples = input_shape[0]
        return tf.TensorShape((n_samples, self.n_units))

    def create_variables(self, input_shape):
        if self.n_units is None:
            return

        input_shape = tf.TensorShape(input_shape)
        self.input_shape = input_shape
        _, n_input_features = input_shape

        if n_input_features.value is None:
            raise WeightInitializationError(
                "Cannot create variables for the layer `{}`, because "
                "number of input features is unknown. Input shape: {}"
                "Layer: {}".format(self.name, input_shape, self))

        self.weight = self.variable(
            value=self.weight, name='weight',
            shape=as_tuple(n_input_features, self.n_units))

        if self.bias is not None:
            self.bias = self.variable(
                value=self.bias, name='bias',
                shape=as_tuple(self.n_units))

    def output(self, input, **kwargs):
        input = tf.convert_to_tensor(input, dtype=tf.float32)

        if self.n_units is None:
            return self.activation_function(input)

        if self.bias is None:
            output = tf.matmul(input, self.weight)
            return self.activation_function(output)

        output = tf.matmul(input, self.weight) + self.bias
        return self.activation_function(output)

    def activation_function(self, input_value):
        return input_value

    def __repr__(self):
        if self.n_units is None:
            return self._repr_arguments(name=self.name)

        return self._repr_arguments(
            self.n_units,
            name=self.name,
            weight=self.weight,
            bias=self.bias,
        )


class Sigmoid(Linear):
    """
    Layer with the sigmoid used as an activation function. It applies
    linear transformation when the ``n_units`` parameter specified and
    sigmoid function after the transformation. When ``n_units`` is not
    specified, only sigmoid function will be applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Logistic Regression (LR)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Sigmoid(1)

    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Sigmoid(5) >> Sigmoid(1)

    Convolutional Neural Networks (CNN) for Semantic Segmentation

    Sigmoid layer can be used in order to normalize probabilities
    per pixel in semantic classification task with two classes.
    In the example below, we have as input 32x32 image that predicts
    one of the two classes. Sigmoid normalizes raw predictions per pixel
    to the valid probabilities.

    >>> from neupy.layers import *
    >>> network = Input((32, 32, 1)) >> Sigmoid()
    """
    def activation_function(self, input_value):
        return tf.nn.sigmoid(input_value)


class HardSigmoid(Linear):
    """
    Layer with the hard sigmoid used as an activation function. It applies
    linear transformation when the ``n_units`` parameter specified and
    hard sigmoid function after the transformation. When ``n_units`` is
    not specified, only hard sigmoid function will be applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> HardSigmoid(5)
    """
    def activation_function(self, input_value):
        input_value = (0.2 * input_value) + 0.5
        return tf.clip_by_value(input_value, 0., 1.)


class Tanh(Linear):
    """
    Layer with the hyperbolic tangent used as an activation function.
    It applies linear transformation when the ``n_units`` parameter
    specified and ``tanh`` function after the transformation. When
    ``n_units`` is not specified, only ``tanh`` function will be applied
    to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Tanh(5)
    """
    def activation_function(self, input_value):
        return tf.nn.tanh(input_value)


class Relu(Linear):
    """
    Layer with the rectifier (ReLu) used as an activation function.
    It applies linear transformation when the ``n_units`` parameter
    specified and ``relu`` function after the transformation. When
    ``n_units`` is not specified, only ``relu`` function will be applied
    to the input.

    Parameters
    ----------
    {Linear.n_units}

    alpha : float
        Alpha parameter defines the decreasing rate
        for the negative values. If ``alpha``
        is non-zero value then layer behave like a
        leaky ReLu. Defaults to ``0``.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`HeNormal(gain=2) <neupy.init.HeNormal>`.

    {Linear.bias}

    {BaseLayer.name}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Relu(20) >> Relu(1)

    Convolutional Neural Networks (CNN)

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((32, 32, 3)),
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     Convolution((3, 3, 32)) >> Relu(),
    ...     Reshape(),
    ...     Softmax(10),
    ... )
    """
    alpha = NumberProperty(minval=0)

    def __init__(self, n_units=None, alpha=0, weight=init.HeNormal(gain=2),
                 bias=init.Constant(value=0), name=None):

        self.alpha = alpha
        super(Relu, self).__init__(
            n_units=n_units, weight=weight, bias=bias, name=name)

    def activation_function(self, input_value):
        if self.alpha == 0:
            return tf.nn.relu(input_value)
        return tf.nn.leaky_relu(input_value, asfloat(self.alpha))

    def __repr__(self):
        if self.n_units is None:
            return self._repr_arguments(name=self.name, alpha=self.alpha)

        return self._repr_arguments(
            self.n_units,
            name=self.name,
            alpha=self.alpha,
            weight=self.weight,
            bias=self.bias,
        )


class LeakyRelu(Linear):
    """
    Layer with the leaky rectifier (Leaky ReLu) used as an activation
    function. It applies linear transformation when the ``n_units``
    parameter specified and leaky relu function after the transformation.
    When ``n_units`` is not specified, only leaky relu function will be
    applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Notes
    -----
    Do the same as ``Relu(input_size, alpha=0.01)``.

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> LeakyRelu(20) >> LeakyRelu(1)
    """
    def activation_function(self, input_value):
        return tf.nn.leaky_relu(input_value, alpha=asfloat(0.01))


class Softplus(Linear):
    """
    Layer with the softplus used as an activation function. It applies linear
    transformation when the ``n_units`` parameter specified and softplus
    function after the transformation. When ``n_units`` is not specified,
    only softplus function will be applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Softplus(4)
    """
    def activation_function(self, input_value):
        return tf.nn.softplus(input_value)


class Softmax(Linear):
    """
    Layer with the softmax activation function. It applies linear
    transformation when the ``n_units`` parameter specified and softmax
    function after the transformation. When ``n_units`` is not specified,
    only softmax function will be applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Relu(20) >> Softmax(10)

    Convolutional Neural Networks (CNN) for Semantic Segmentation

    Softmax layer can be used in order to normalize probabilities
    per pixel. In the example below, we have as input 32x32 image
    with raw prediction per each pixel for 10 different classes.
    Softmax normalizes raw predictions per pixel to the probability
    distribution.

    >>> from neupy.layers import *
    >>> network = Input((32, 32, 10)) >> Softmax()
    """
    def activation_function(self, input_value):
        return tf.nn.softmax(input_value)


class Elu(Linear):
    """
    Layer with the exponential linear unit (ELU) used as an activation
    function. It applies linear transformation when the ``n_units``
    parameter specified and elu function after the transformation.
    When ``n_units`` is not specified, only elu function will be
    applied to the input.

    Parameters
    ----------
    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> Elu(5) >> Elu(1)

    References
    ----------
    .. [1] http://arxiv.org/pdf/1511.07289v3.pdf
    """
    def activation_function(self, input_value):
        return tf.nn.elu(input_value)


class PRelu(Linear):
    """
    Layer with the parametrized ReLu used as an activation function.
    Layer learns additional parameter ``alpha`` during the training.

    It applies linear transformation when the ``n_units`` parameter
    specified and parametrized relu function after the transformation.
    When ``n_units`` is not specified, only parametrized relu function
    will be applied to the input.

    Parameters
    ----------
    alpha_axes : int or tuple
        Axes that will not include unique alpha parameter.
        Single integer value defines the same as a tuple with one value.
        Defaults to ``-1``.

    alpha : array-like, Tensorfow variable, scalar or Initializer
        Separate alpha parameter per each non-shared axis for the ReLu.
        Scalar value means that each element in the tensor will be
        equal to the specified value. Default initialization methods you
        can find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0.25)``.

    {Linear.Parameters}

    Methods
    -------
    {Linear.Methods}

    Attributes
    ----------
    {Linear.Attributes}

    Examples
    --------
    Feedforward Neural Networks (FNN)

    >>> from neupy.layers import *
    >>> network = Input(10) >> PRelu(20) >> PRelu(1)

    Convolutional Neural Networks (CNN)

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((32, 32, 3)),
    ...     Convolution((3, 3, 16)) >> PRelu(),
    ...     Convolution((3, 3, 32)) >> PRelu(),
    ...     Reshape(),
    ...     Softmax(10),
    ... )

    References
    ----------
    .. [1] Delving Deep into Rectifiers: Surpassing Human-Level
           Performance on ImageNet Classification.
           https://arxiv.org/pdf/1502.01852v1.pdf
    """
    alpha_axes = TypedListProperty()
    alpha = ParameterProperty()

    def __init__(self, n_units=None, alpha_axes=-1, alpha=0.25,
                 weight=init.HeNormal(gain=2), bias=0, name=None):

        self.alpha = alpha
        self.alpha_axes = as_tuple(alpha_axes)

        if 0 in self.alpha_axes:
            raise ValueError("Cannot specify alpha for 0-axis")

        super(PRelu, self).__init__(
            n_units=n_units, weight=weight, bias=bias, name=name)

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if input_shape and max(self.alpha_axes) >= input_shape.ndims:
            max_axis_index = input_shape.ndims - 1

            raise LayerConnectionError(
                "Cannot specify alpha for the axis #{}. Maximum "
                "available axis is {} (0-based indices)."
                "".format(max(self.alpha_axes), max_axis_index))

        return super(PRelu, self).get_output_shape(input_shape)

    def create_variables(self, input_shape):
        super(PRelu, self).create_variables(input_shape)
        output_shape = self.get_output_shape(input_shape)

        self.alpha = self.variable(
            value=self.alpha, name='alpha',
            shape=[output_shape[axis] for axis in self.alpha_axes])

    def activation_function(self, input):
        input = tf.convert_to_tensor(input, dtype=tf.float32)
        ndim = input.shape.ndims

        dimensions = np.arange(ndim)
        alpha_axes = dimensions[list(self.alpha_axes)]

        alpha = tf_utils.dimshuffle(self.alpha, ndim, alpha_axes)
        return tf.maximum(0.0, input) + alpha * tf.minimum(0.0, input)

    def __repr__(self):
        if self.n_units is None:
            return self._repr_arguments(
                name=self.name,
                alpha_axes=self.alpha_axes,
                alpha=self.alpha)

        return self._repr_arguments(
            self.n_units,
            name=self.name,
            alpha_axes=self.alpha_axes,
            alpha=self.alpha,
            weight=self.weight,
            bias=self.bias)



import tensorflow as tf

from neupy.core.config import Configurable
from neupy.core.properties import (
    ChoiceProperty, NumberProperty,
    WithdrawProperty, IntProperty,
)
from neupy.utils.tf_utils import setup_parameter_updates
from neupy.algorithms.minsearch.wolfe import line_search
from neupy.utils import (
    asfloat, dot, outer, as_tuple,
    function_name_scope, make_single_vector,
)
from .base import BaseOptimizer


__all__ = ('QuasiNewton',)


class WolfeLineSearchForStep(Configurable):
    """
    Class that has all functions required in order to apply line search over
    step parameter that used during the network training.

    Parameters
    ----------
    wolfe_maxiter : int
        Controls maximum number of iteration during the line search that
        identifies optimal step size during the weight update stage.
        Defaults to ``20``.

    wolfe_c1 : float
        Parameter for Armijo condition rule. It's used during the line search
        that identifies optimal step size during the weight update stage.
        Defaults ``1e-4``.

    wolfe_c2 : float
        Parameter for curvature condition rule. It's used during the line
        search that identifies optimal step size during the weight update
        stage. Defaults ``0.9``.
    """
    wolfe_maxiter = IntProperty(default=20, minval=0)
    wolfe_c1 = NumberProperty(default=1e-4, minval=0)
    wolfe_c2 = NumberProperty(default=0.9, minval=0)

    def find_optimal_step(self, parameter_vector, parameter_update):
        network_inputs = as_tuple(self.network.inputs)
        layers_and_parameters = list(self.network.variables.items())

        def prediction(step):
            step = asfloat(step)
            updated_params = parameter_vector + step * parameter_update

            try:
                # This trick allow us to replace shared variables
                # with tensorflow variables and get output from the network
                start_pos = 0
                for (layer, varname), variable in layers_and_parameters:
                    n_param_values = int(variable.shape.num_elements())
                    end_pos = start_pos + n_param_values

                    updated_param_value = tf.reshape(
                        updated_params[start_pos:end_pos],
                        variable.shape)

                    setattr(layer, varname, updated_param_value)
                    start_pos = end_pos

                output = self.network.output(*network_inputs, training=True)

            finally:
                # Restore previous parameters
                for (layer, varname), param in layers_and_parameters:
                    setattr(layer, varname, param)

            return output

        def phi(step):
            return self.loss(self.target, prediction(step))

        def derphi(step):
            error_func = self.loss(self.target, prediction(step))
            gradient, = tf.gradients(error_func, step)
            return gradient

        return line_search(
            phi, derphi, self.wolfe_maxiter,
            self.wolfe_c1, self.wolfe_c2)


@function_name_scope
def safe_reciprocal(value, epsilon):
    """
    The same as regular function in the tensorflow accept that it ensures
    that non of the input values have magnitude smaller than epsilon.
    Otherwise small values will be capped to the epsilon.
    """
    inv_epsilon = 1. / epsilon
    return tf.clip_by_value(
        tf.reciprocal(value),
        -inv_epsilon,
        inv_epsilon
    )


@function_name_scope
def safe_division(numerator, denominator, epsilon):
    """
    The same as regular function in the tensorflow accept that it ensures
    that non of the denominator values have magnutide smaller than epsilon.
    Otherwise small values will be capped to the epsilon.
    """
    inv_denominator = safe_reciprocal(denominator, epsilon)
    return numerator * inv_denominator


@function_name_scope
def bfgs(inv_H, delta_w, delta_grad, epsilon=1e-7):
    """
    It can suffer from round-off error and inaccurate line searches.
    """
    n_parameters = int(inv_H.shape[0])

    I = tf.eye(n_parameters)
    rho = safe_reciprocal(dot(delta_grad, delta_w), epsilon)

    X = I - outer(delta_w, delta_grad) * rho
    X_T = tf.transpose(X)
    Z = rho * outer(delta_w, delta_w)

    return tf.matmul(X, tf.matmul(inv_H, X_T)) + Z


@function_name_scope
def dfp(inv_H, delta_w, delta_grad, epsilon=1e-7):
    """
    DFP is a method very similar to BFGS. It's rank 2 formula update.
    It can suffer from round-off error and inaccurate line searches.
    """
    inv_H_dot_grad = dot(inv_H, delta_grad)

    x = safe_division(
        outer(delta_w, delta_w),
        dot(delta_grad, delta_w),
        epsilon
    )
    y = safe_division(
        tf.matmul(outer(inv_H_dot_grad, delta_grad), inv_H),
        dot(delta_grad, inv_H_dot_grad),
        epsilon
    )

    return inv_H - y + x


@function_name_scope
def sr1(inv_H, delta_w, delta_grad, epsilon=1e-7):
    """
    Symmetric rank 1 (SR1). Generates update for the inverse hessian
    matrix adding symmetric rank-1 matrix. It's possible that there is no
    rank 1 updates for the matrix and in this case update won't be applied
    and original inverse hessian will be returned.
    """
    param = delta_w - dot(inv_H, delta_grad)
    denominator = dot(param, delta_grad)

    return tf.where(
        # This check protects from the cases when update
        # doesn't exist. It's possible that during certain
        # iteration there is no rank-1 update for the matrix.
        tf.less(
            tf.abs(denominator),
            epsilon * tf.norm(param) * tf.norm(delta_grad)
        ),
        inv_H,
        inv_H + outer(param, param) / denominator
    )


class QuasiNewton(WolfeLineSearchForStep, BaseOptimizer):
    """
    Quasi-Newton algorithm. Every iteration quasi-Network method approximates
    inverse Hessian matrix with iterative updates. It doesn't have ``step``
    parameter. Instead, algorithm applies line search for the step parameter
    that satisfies strong Wolfe condition. Parameters that control wolfe
    search start with the ``wolfe_`` prefix.

    Parameters
    ----------
    update_function : ``bfgs``, ``dfp``, ``sr1``
        Update function for the iterative inverse hessian matrix
        approximation. Defaults to ``bfgs``.

        - ``bfgs`` -  It's rank 2 formula update. It can suffer from
          round-off error and inaccurate line searches.

        - ``dfp`` - DFP is a method very similar to BFGS. It's rank 2 formula
          update. It can suffer from round-off error and inaccurate line
          searches.

        - ``sr1`` - Symmetric rank 1 (SR1). Generates update for the
          inverse hessian matrix adding symmetric rank-1 matrix. It's
          possible that there is no rank 1 updates for the matrix and in
          this case update won't be applied and original inverse hessian
          will be returned.

    h0_scale : float
        Default Hessian matrix is an identity matrix. The
        ``h0_scale`` parameter scales identity matrix.
        Defaults to ``1``.

    epsilon : float
        Controls numerical stability for the ``update_function`` parameter.
        Defaults to ``1e-7``.

    {WolfeLineSearchForStep.Parameters}

    {BaseOptimizer.network}

    {BaseOptimizer.loss}

    {BaseOptimizer.show_epoch}

    {BaseOptimizer.shuffle_data}

    {BaseOptimizer.signals}

    {BaseOptimizer.verbose}

    {BaseOptimizer.regularizer}

    Notes
    -----
    - Method requires all training data during propagation, which means
      it cannot be trained with mini-batches.

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> optimizer = algorithms.QuasiNewton(
    ...     Input(2) >> Sigmoid(3) >> Sigmoid(1),
    ...     update_function='bfgs'
    ... )
    >>> optimizer.train(x_train, y_train, epochs=10)

    References
    ----------
    [1] Yang Ding, Enkeleida Lushi, Qingguo Li,
        Investigation of quasi-Newton methods for unconstrained optimization.
        http://people.math.sfu.ca/~elushi/project_833.pdf

    [2] Jorge Nocedal, Stephen J. Wright, Numerical Optimization.
        Chapter 6, Quasi-Newton Methods, p. 135-163
    """
    update_function = ChoiceProperty(
        default='bfgs',
        choices={
            'bfgs': bfgs,
            'dfp': dfp,
            'sr1': sr1,
        }
    )
    epsilon = NumberProperty(default=1e-7, minval=0)
    h0_scale = NumberProperty(default=1, minval=0)
    step = WithdrawProperty()

    def init_variables(self):
        with tf.name_scope('quasi-newton'):
            n_parameters = self.network.n_parameters
            self.variables.update(
                inv_hessian=tf.Variable(
                    asfloat(self.h0_scale) * tf.eye(n_parameters),
                    name="inv-hessian",
                    dtype=tf.float32,
                ),
                prev_params=tf.Variable(
                    tf.zeros([n_parameters]),
                    name="prev-params",
                    dtype=tf.float32,
                ),
                prev_full_gradient=tf.Variable(
                    tf.zeros([n_parameters]),
                    name="prev-full-gradient",
                    dtype=tf.float32,
                ),
                iteration=tf.Variable(
                    asfloat(self.last_epoch),
                    name='current-iteration',
                    dtype=tf.float32
                ),
            )

    def init_train_updates(self):
        self.init_variables()

        iteration = self.variables.iteration
        inv_hessian = self.variables.inv_hessian
        prev_params = self.variables.prev_params
        prev_full_gradient = self.variables.prev_full_gradient

        variables = self.network.variables
        params = [var for var in variables.values() if var.trainable]
        param_vector = make_single_vector(params)

        gradients = tf.gradients(self.variables.loss, params)
        full_gradient = make_single_vector(gradients)

        new_inv_hessian = tf.where(
            tf.equal(iteration, 0),
            inv_hessian,
            self.update_function(
                inv_H=inv_hessian,
                delta_w=param_vector - prev_params,
                delta_grad=full_gradient - prev_full_gradient,
                epsilon=self.epsilon
            )
        )
        param_delta = -dot(new_inv_hessian, full_gradient)
        step = self.find_optimal_step(param_vector, param_delta)
        updated_params = param_vector + step * param_delta
        updates = setup_parameter_updates(params, updated_params)

        # We have to compute these values first, otherwise
        # parallelization, in tensorflow, can mix update order
        # and, for example, previous gradient can be equal to
        # current gradient value. It happens because tensorflow
        # try to execute operations in parallel.
        required_variables = [new_inv_hessian, param_vector, full_gradient]
        with tf.control_dependencies(required_variables):
            updates.extend([
                inv_hessian.assign(new_inv_hessian),
                prev_params.assign(param_vector),
                prev_full_gradient.assign(full_gradient),
                iteration.assign(iteration + 1),
            ])

        return updates



from functools import wraps

import numpy as np
import tensorflow as tf

from neupy.utils.misc import as_tuple
from neupy.utils.processing import asfloat


__all__ = (
    # Main tensorflow functions
    'tensorflow_session', 'tensorflow_eval', 'create_variable',
    'initialize_uninitialized_variables', 'function',

    # Functions that help to deal with tensorflow name scope
    'class_method_name_scope', 'function_name_scope',

    # Misc utils for tensorflow
    'flatten', 'outer', 'repeat', 'dimshuffle', 'shape_to_tuple',
    'dot', 'make_single_vector', 'setup_parameter_updates',
)


def function(inputs, outputs, updates=None, name=None):
    """
    Function simulates behaviour of the Theano's functions.

    Parameters
    ----------
    inputs : list
        List of input placeholders

    outputs : list, Tensor
        Output that has to be computed by the function

    updates : list or None
        List of the updates that has to be performed on the variables.
        The ``None`` value means that no updates will be applied at the
        end of the computation. Defaults to ``None``.

    name : str or None
        Defaults to ``None``.

    Returns
    -------
    function
    """
    if updates is None:
        updates = []

    session = tensorflow_session()
    tensorflow_updates = []

    # Ensure that all new values has been computed. Absence of these
    # checks might lead to the non-deterministic update behaviour.
    new_values = [val[1] for val in updates if isinstance(val, (list, tuple))]

    # Make sure that all outputs has been computed
    with tf.control_dependencies(as_tuple(outputs, new_values)):
        for update in updates:
            if isinstance(update, (list, tuple)):
                old_value, new_value = update
                update = old_value.assign(new_value)
            tensorflow_updates.append(update)

        # Group variables in order to avoid output for the updates
        tensorflow_updates = tf.group(*tensorflow_updates)

    @wraps(function)
    def wrapper(*input_values):
        feed_dict = dict(zip(inputs, input_values))
        result, _ = session.run(
            [outputs, tensorflow_updates],
            feed_dict=feed_dict,
        )
        return result
    return wrapper


def tensorflow_session():
    if hasattr(tensorflow_session, 'cache'):
        session = tensorflow_session.cache

        if not session._closed:
            return session

    config = tf.ConfigProto(
        allow_soft_placement=True,
        inter_op_parallelism_threads=0,
        intra_op_parallelism_threads=0,
    )
    session = tf.Session(config=config)

    tensorflow_session.cache = session
    return session


def initialize_uninitialized_variables(variables=None):
    if variables is None:
        variables = tf.global_variables()

    if not variables:
        return

    session = tensorflow_session()
    is_not_initialized = session.run([
        tf.is_variable_initialized(var) for var in variables])

    not_initialized_vars = [
        v for (v, f) in zip(variables, is_not_initialized) if not f]

    if len(not_initialized_vars):
        session.run(tf.variables_initializer(not_initialized_vars))


def function_name_scope(function):
    """
    Decorator that wraps any function with the name score that has the
    same name as a function.
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        with tf.name_scope(function.__name__):
            return function(*args, **kwargs)
    return wrapper


def class_method_name_scope(method):
    """
    Decorator that wraps any method with the name score that has the
    same name as a method.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with tf.name_scope(self.__class__.__name__):
            if hasattr(method, '__self__'):  # check if method bounded
                return method(*args, **kwargs)
            return method(self, *args, **kwargs)

    wrapper.original_method = method
    return wrapper


def tensorflow_eval(value):
    session = tensorflow_session()
    initialize_uninitialized_variables()
    return session.run(value)


@function_name_scope
def flatten(value):
    return tf.reshape(value, [-1])


@function_name_scope
def outer(a, b):
    a = tf.expand_dims(a, 1)  # column vector
    b = tf.expand_dims(b, 0)  # row vector
    return tf.matmul(a, b)


@function_name_scope
def dot(a, b):
    return tf.tensordot(a, b, 1)


def repeat(tensor, repeats):
    """
    Repeat elements of an tensor. The same as ``numpy.repeat``.

    Parameters
    ----------
    input : tensor
    repeats: list, tuple
        Number of repeat for each dimension, length must be the
        same as the number of dimensions in input.

    Returns
    -------
    tensor
        Has the same type as input. Has the shape
        of ``tensor.shape * repeats``.
    """
    with tf.variable_scope("repeat"):
        expanded_tensor = tf.expand_dims(tensor, -1)

        multiples = as_tuple(1, repeats)
        tiled_tensor = tf.tile(expanded_tensor, multiples)

        repeats = tf.convert_to_tensor(repeats)
        return tf.reshape(tiled_tensor, tf.shape(tensor) * repeats)


def make_single_vector(parameters):
    with tf.name_scope('parameters-vector'):
        return tf.concat([flatten(param) for param in parameters], axis=0)


def setup_parameter_updates(parameters, parameter_update_vector):
    """
    Creates update rules for list of parameters from one vector.
    Function is useful in Conjugate Gradient or
    Levenberg-Marquardt optimization algorithms

    Parameters
    ----------
    parameters : list
        List of parameters.

    parameter_update_vector : Tensorfow varible
        Vector that contains updates for all parameters.

    Returns
    -------
    list
        List of updates separeted for each parameter.
    """
    updates = []
    start_position = 0

    for parameter in parameters:
        end_position = start_position + tf.size(parameter)
        new_parameter = tf.reshape(
            parameter_update_vector[start_position:end_position],
            parameter.shape
        )
        updates.append((parameter, new_parameter))
        start_position = end_position

    return updates


def dimshuffle(value, ndim, axes):
    """
    Shuffle dimension based on the specified number of
    dimensions and axes.

    Parameters
    ----------
    value : Tensorfow variable
    ndim : int
    axes : tuple, list

    Returns
    -------
    Tensorfow variable
    """
    for dim in range(ndim):
        if dim not in axes:
            value = tf.expand_dims(value, dim)
    return value


def shape_to_tuple(shape):
    if isinstance(shape, tf.TensorShape):
        if shape.ndims is not None:
            return tuple([dim.value for dim in shape.dims])
        return None

    if isinstance(shape, tf.Dimension):
        return shape.value

    if isinstance(shape, list):
        return [shape_to_tuple(s) for s in shape]

    if isinstance(shape, tuple):
        return tuple([shape_to_tuple(s) for s in shape])

    return shape


def create_variable(value, name, shape, trainable=True):
    """
    Creates NN parameter as Tensorfow variable.

    Parameters
    ----------
    value : array-like, Tensorfow variable, scalar or Initializer
        Default value for the parameter.

    name : str
        Shared variable name.

    shape : tuple
        Parameter's shape.

    trainable : bool
        Whether parameter trainable by backpropagation.

    Returns
    -------
    Tensorfow variable.
    """
    from neupy import init

    if shape is not None:
        shape = shape_to_tuple(shape)

    if isinstance(value, (tf.Variable, tf.Tensor, np.ndarray, np.matrix)):
        variable_shape = shape_to_tuple(value.shape)

        if as_tuple(variable_shape) != as_tuple(shape):
            raise ValueError(
                "Cannot create variable with name `{}`. Provided variable "
                "with shape {} is incompatible with expected shape {}"
                "".format(name, variable_shape, shape))

    if isinstance(value, (tf.Variable, tf.Tensor)):
        return value

    if isinstance(value, (int, float)):
        value = init.Constant(value)

    if isinstance(value, init.Initializer):
        value = value.sample(shape)

    return tf.Variable(
        asfloat(value),
        name=name,
        dtype=tf.float32,
        trainable=trainable,
    )



import matplotlib.pyplot as plt
from sklearn import datasets
from neupy import algorithms, utils

from helpers import plot_2d_grid


plt.style.use('ggplot')
utils.reproducible()


if __name__ == '__main__':
    GRID_WIDTH = 20
    GRID_HEIGHT = 1

    data, targets = datasets.make_moons(n_samples=400, noise=0.1)
    data = data[targets == 1]

    sofm = algorithms.SOFM(
        n_inputs=2,
        features_grid=(GRID_HEIGHT, GRID_WIDTH),

        verbose=True,
        shuffle_data=True,

        # The winning neuron will be selected based on the
        # Euclidean distance. For this task it's important
        # that distance is Euclidean. Other distances will
        # not give us the same results.
        distance='euclid',

        learning_radius=2,
        # Reduce learning radius by 1 after every 20 epochs.
        # Learning radius will be equal to 2 during first
        # 20 epochs and on the 21st epoch it will be equal to 1.
        reduce_radius_after=20,

        # 2 Means that neighbour neurons will have high learning
        # rates during the first iterations
        std=2,
        # Defines a rate at which parameter `std` will be reduced.
        # Reduction is monotonic and reduces after each epoch.
        # In 50 epochs std = 2 / 2 = 1 and after 100 epochs
        # std = 2 / 3 and so on.
        reduce_std_after=50,

        # Step (or learning rate)
        step=0.3,
        # Defines a rate at which parameter `step` will reduced.
        # Reduction is monotonic and reduces after each epoch.
        # In 50 epochs step = 0.3 / 2 = 0.15 and after 100 epochs
        # std = 0.3 / 3 = 0.1 and so on.
        reduce_step_after=50,
    )
    sofm.train(data, epochs=20)

    red, blue = ('#E24A33', '#348ABD')

    plt.scatter(*data.T, color=blue)
    plt.scatter(*sofm.weight, color=red)

    weights = sofm.weight.reshape((2, GRID_HEIGHT, GRID_WIDTH))
    plot_2d_grid(weights, color=red)

    plt.show()



from abc import ABCMeta
from collections import namedtuple

import numpy as np
import tensorflow as tf
from six import with_metaclass

from neupy.utils import tensorflow_eval
from .properties import BaseProperty, WithdrawProperty
from .docs import SharedDocsMeta


__all__ = ('ConfigMeta', 'ConfigABCMeta', 'Configurable', 'ConfigurableABC',
           'ExtractParameters', 'DumpableObject')


Option = namedtuple('Option', 'class_name value')


class ExtractParameters(object):
    def get_params(self, deep=False):
        options = {}

        for property_name, option in self.options.items():
            value = getattr(self, property_name)

            if isinstance(value, tf.Variable):
                value = tensorflow_eval(value)

            property_ = option.value
            is_numpy_array = isinstance(value, np.ndarray)

            if hasattr(option.value, 'choices'):
                choices = property_.choices

                if not is_numpy_array and value in choices.values():
                    choices = {v: k for k, v in choices.items()}
                    value = choices[value]

            options[property_name] = value

        return options

    def set_params(self, **params):
        self.__dict__.update(params)
        return self


def initialize_with_kwargs(class_, kwargs):
    return class_(**kwargs)


class DumpableObject(ExtractParameters):
    def __reduce__(self):
        return initialize_with_kwargs, (self.__class__, self.get_params())


class ConfigMeta(SharedDocsMeta):
    """
    Meta-class that configure initialized properties. Also it helps
    inheit properties from parent classes and use them.
    """
    def __new__(cls, clsname, bases, attrs):
        new_class = super(ConfigMeta, cls).__new__(cls, clsname, bases, attrs)
        parents = [kls for kls in bases if isinstance(kls, ConfigMeta)]

        if not hasattr(new_class, 'options'):
            new_class.options = {}

        for base_class in parents:
            new_class.options = dict(base_class.options, **new_class.options)

        options = new_class.options

        # Set properties names and save options for different classes
        for key, value in attrs.items():
            if isinstance(value, BaseProperty):
                value.name = key
                options[key] = Option(class_name=clsname, value=value)

            if isinstance(value, WithdrawProperty) and key in options:
                del options[key]

        return new_class


class BaseConfigurable(ExtractParameters):
    """
    Base configuration class. It help set up and validate
    initialized property values.

    Parameters
    ----------
    **options
        Available properties.
    """
    def __init__(self, **options):
        available_options = set(self.options.keys())
        invalid_options = set(options) - available_options

        if invalid_options:
            clsname = self.__class__.__name__
            raise ValueError(
                "The `{}` object contains invalid properties: {}"
                "".format(clsname, ', '.join(invalid_options)))

        for key, value in options.items():
            setattr(self, key, value)

        for option_name, option in self.options.items():
            if option.value.required and option_name not in options:
                raise ValueError(
                    "Option `{}` is required.".format(option_name))


class Configurable(with_metaclass(ConfigMeta, BaseConfigurable)):
    """
    Class that combine ``BaseConfigurable`` class functionality and
    ``ConfigMeta`` meta-class.
    """


class ConfigABCMeta(ABCMeta, ConfigMeta):
    """
    Meta-class that combines ``ConfigMeta`` and ``abc.ABCMeta``
    meta-classes.
    """


class ConfigurableABC(with_metaclass(ConfigABCMeta, BaseConfigurable)):
    """
    Class that combine ``BaseConfigurable`` class functionality,
    ``ConfigMeta`` and ``abc.ABCMeta`` meta-classes.
    """



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import sys

from .config import Configurable
from .properties import BaseProperty


__all__ = ('Verbose',)


def is_color_supported():
    """
    Returns ``True`` if the running system's terminal supports
    color, and ``False`` otherwise.

    Notes
    -----
    Code from Djano: https://github.com/django/django/blob/\
    master/django/core/management/color.py

    Returns
    -------
    bool
    """
    supported_platform = (
        sys.platform != 'Pocket PC' and
        (sys.platform != 'win32' or 'ANSICON' in os.environ)
    )

    # isatty is not always implemented
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    is_support = (supported_platform and is_a_tty)
    return is_support


def create_style(ansi_code, use_bright_mode=False):
    """
    Create style based on ANSI code number.

    Parameters
    ----------
    ansi_code : int
        ANSI style code.

    Returns
    -------
    function
        Function that takes string argument and add ANDI styles
        if its possible.
    """
    def style(text):
        if is_color_supported():
            mode = int(use_bright_mode)
            return "\033[{};{}m{}\033[0m".format(mode, ansi_code, text)
        return text
    return style


class TerminalLogger(object):
    """
    Customized logging class that replace standard logging
    functionality.

    Attributes
    ----------
    enable : bool
        Enable/disable logging output. Defaults to ``True``.

    template : str
        Terminal output message template. Defaults
        to ``"[{name}] {text}"``.

    stdout : object
        Writes output in terminal. Defaults to ``sys.stdout``.
    """
    colors = {
        'gray': create_style(ansi_code=37),
        'green': create_style(ansi_code=32),
        'red': create_style(ansi_code=31),
        'white': create_style(ansi_code=37),
    }
    styles = {
        'bold': create_style(ansi_code=1, use_bright_mode=True),
        'underline': create_style(ansi_code=4, use_bright_mode=True),
    }

    def __init__(self, enable=True):
        self.enable = enable
        self.template = "[{tag}] {text}"
        self.stdout = sys.stdout

    def write(self, text):
        """
        Method writes text in terminal if logging is enable.

        Parameters
        ----------
        text : str
        """
        if self.enable:
            self.stdout.write(text)
            self.stdout.write('\n')

    def newline(self):
        self.write('\r')

    def message(self, tag, text, color='green'):
        """
        Methods writes message in terminal using specific template.
        Each row should have tag and text. Tag identifies message
        category and text information related to this category.

        Parameters
        ----------
        name : str
        text : str
        color : {{'green', 'gray', 'red', 'white'}}
            Property that color text defined as ``tag`` parameter.
            Defaults to ``green``.
        """
        if color not in self.colors:
            available_colors = ', '.join(self.colors.keys())
            raise ValueError(
                "Invalid color `{}`. Available colors: {}"
                "".format(color, available_colors))

        colorizer = self.colors[color]
        formated_tag = colorizer(tag.upper())
        message = self.template.format(tag=formated_tag, text=text)
        self.write(message)

    def title(self, text):
        """
        Method write text as a title message. Text will be displayed
        using bold and underline text styles. Also there will be empty
        lines before and after the message.

        Parameters
        ----------
        text : str
        """
        bold = self.styles['bold']
        underline = self.styles['underline']
        self.write("\n{text}\n".format(text=underline(bold(text))))

    def __reduce__(self):
        return (self.__class__, (self.enable,))


class VerboseProperty(BaseProperty):
    """
    Property that synchronize updates with ``enable`` attribute in
    logging instance.

    Parameters
    ----------
    {BaseProperty.default}
    {BaseProperty.required}
    """
    expected_type = bool

    def __set__(self, instance, value):
        instance.logs.enable = value
        return super(VerboseProperty, self).__set__(instance, value)


class Verbose(Configurable):
    """
    Class that controls NeuPy logging.

    Parameters
    ----------
    verbose : bool
        Property controls verbose output in terminal. The ``True`` value
        enables informative output in the terminal and ``False`` - disable
        it. Defaults to ``False``.

    Attributes
    ----------
    logs : TerminalLogger
        ``TerminalLogger`` instance.
    """
    verbose = VerboseProperty(default=False)

    def __init__(self, **options):
        self.logs = TerminalLogger()
        self.logs.enable = self.verbose
        super(Verbose, self).__init__(**options)



import numpy as np

from neupy.utils import format_data, iters
from neupy.core.properties import BoundedProperty, IntProperty
from neupy.algorithms.base import BaseSkeleton
from neupy.exceptions import NotTrained
from .utils import pdf_between_data


__all__ = ('PNN',)


class PNN(BaseSkeleton):
    """
    Probabilistic Neural Network (PNN). Network applies only to
    the classification problems.

    Notes
    -----
    - PNN Network is sensitive for cases when one input feature
      has higher values than the other one. Input data has to be
      normalized before training.

    - Standard deviation has to match the range of the input features
      Check ``std`` parameter description for more information.

    - The bigger training dataset the slower prediction.
      Algorithm is much more efficient for small datasets.

    - Network uses lazy learning which mean that network doesn't
      need iterative training. It just stores parameters
      and use them to make a predictions.

    Parameters
    ----------
    std : float
        Standard deviation for the Probability Density Function (PDF).
        If your input features have high values than standard deviation
        should also be high. For instance, if input features from range
        ``[0, 20]`` that standard deviation should be also a big value
        like ``10`` or ``15``. Small values will lead to bad prediction.

    batch_size : int or None
        Set up min-batch size. The ``None`` value will ensure that all data
        samples will be propagated through the network at once.
        Defaults to ``128``.

    {Verbose.verbose}

    Methods
    -------
    train(X_train, y_train, copy=True)
        Network just stores all the information about the data and use
        it for the prediction. Parameter ``copy`` copies input data
        before saving it inside the network.

        The ``y_train`` argument should be a vector or
        matrix with one feature column.

    predict(X)
        Return classes associated with each sample in the ``X``.

    predict_proba(X)
        Predict probabilities for each class.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from sklearn import datasets, metrics
    >>> from sklearn.model_selection import train_test_split
    >>> from neupy import algorithms
    >>>
    >>> dataset = datasets.load_digits()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     dataset.data, dataset.target, test_size=0.3
    ... )
    >>>
    >>> pnn = algorithms.PNN(std=10, verbose=False)
    >>> pnn.train(x_train, y_train)
    >>>
    >>> y_predicted = pnn.predict(x_test)
    >>> metrics.accuracy_score(y_test, y_predicted)
    0.98888888888888893
    """
    std = BoundedProperty(minval=0)
    batch_size = IntProperty(default=128, minval=0, allow_none=True)

    def __init__(self, std, batch_size=128, verbose=False):
        self.std = std
        self.batch_size = batch_size

        self.classes = None
        self.X_train = None
        self.y_train = None

        super(PNN, self).__init__(batch_size=batch_size, verbose=verbose)

    def train(self, X_train, y_train, copy=True):
        """
        Trains network. PNN doesn't actually train, it just stores
        input data and use it for prediction.

        Parameters
        ----------
        X_train : array-like (n_samples, n_features)

        y_train : array-like (n_samples,)
            Target variable should be vector or matrix
            with one feature column.

        copy : bool
            If value equal to ``True`` than input matrices will
            be copied. Defaults to ``True``.

        Raises
        ------
        ValueError
            In case if something is wrong with input data.
        """
        X_train = format_data(X_train, copy=copy)
        y_train = format_data(y_train, copy=copy, make_float=False)

        self.X_train = X_train
        self.y_train = y_train

        if X_train.shape[0] != y_train.shape[0]:
            raise ValueError(
                "Number of samples in the input and "
                "target datasets are different")

        if y_train.shape[1] != 1:
            raise ValueError(
                "Target value should be vector or "
                "matrix with only one column")

        classes = self.classes = np.unique(y_train)
        n_classes = classes.size
        n_samples = X_train.shape[0]

        class_ratios = self.class_ratios = np.zeros(n_classes)
        row_comb_matrix = self.row_comb_matrix = np.zeros(
            (n_classes, n_samples))

        for i, class_name in enumerate(classes):
            class_name = classes[i]
            class_val_positions = (y_train == class_name)
            row_comb_matrix[i, class_val_positions.ravel()] = 1
            class_ratios[i] = np.sum(class_val_positions)

    def predict_proba(self, X):
        """
        Predict probabilities for each class.

        Parameters
        ----------
        X : array-like (n_samples, n_features)

        Returns
        -------
        array-like (n_samples, n_classes)
        """
        outputs = iters.apply_batches(
            function=self.predict_raw,
            inputs=format_data(X),
            batch_size=self.batch_size,
            show_progressbar=self.logs.enable,
        )
        raw_output = np.concatenate(outputs, axis=1)

        total_output_sum = raw_output.sum(axis=0).reshape((-1, 1))
        return raw_output.T / total_output_sum

    def predict_raw(self, X):
        """
        Raw prediction.

        Parameters
        ----------
        X : array-like (n_samples, n_features)

        Raises
        ------
        NotTrained
            If network hasn't been trained.

        ValueError
            In case if something is wrong with input data.

        Returns
        -------
        array-like (n_samples, n_classes)
        """
        if self.classes is None:
            raise NotTrained(
                "Cannot make a prediction. Network hasn't been trained yet")

        if X.shape[1] != self.X_train.shape[1]:
            raise ValueError(
                "Input data must contain {0} features, got {1}"
                "".format(self.X_train.shape[1],  X.shape[1]))

        class_ratios = self.class_ratios.reshape((-1, 1))
        pdf_outputs = pdf_between_data(self.X_train, X, self.std)

        return np.dot(self.row_comb_matrix, pdf_outputs) / class_ratios

    def predict(self, X):
        """
        Predicts class from the input data.

        Parameters
        ----------
        X : array-like (n_samples, n_features)

        Returns
        -------
        array-like (n_samples,)
        """
        outputs = iters.apply_batches(
            function=self.predict_raw,
            inputs=format_data(X),
            batch_size=self.batch_size,
            show_progressbar=self.logs.enable,
        )

        raw_output = np.concatenate(outputs, axis=1)
        return self.classes[raw_output.argmax(axis=0)]



from __future__ import division

import math
import collections

import six
import tensorflow as tf

from neupy import init
from neupy.utils import as_tuple
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import (
    TypedListProperty, Property,
    ParameterProperty,
)
from .base import BaseLayer


__all__ = ('Convolution', 'Deconvolution')


class Spatial2DProperty(TypedListProperty):
    expected_type = (list, tuple, int)

    def __init__(self, *args, **kwargs):
        kwargs['element_type'] = int
        super(Spatial2DProperty, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if isinstance(value, collections.Iterable) and len(value) == 1:
            value = (value[0], 1)

        if isinstance(value, int):
            value = (value, value)

        super(Spatial2DProperty, self).__set__(instance, value)

    def validate(self, value):
        super(Spatial2DProperty, self).validate(value)

        if len(value) > 2:
            raise ValueError(
                "Stride can have only one or two elements "
                "in the list. Got {}".format(len(value)))

        if any(element <= 0 for element in value):
            raise ValueError(
                "Stride size should contain only values greater than zero")


def deconv_output_shape(dimension_size, filter_size, padding, stride,
                        dilation=1):
    """
    Computes deconvolution's output shape for one spatial dimension.

    Parameters
    ----------
    dimension_size : int or None
        Size of the dimension. Typically it's image's weight or height.
        It might be equal to ``None`` when we input might have variable
        dimension.

    filter_size : int
        Size of the convolution filter.

    padding : {``valid``, ``same``} or int
        Type or size of the zero-padding.

    stride : int
        Stride size.

    dilation : int
        Dilation rate. Only ``dilation=1`` is supported for the
        deconvolution.

    Returns
    -------
    int
        Dimension size after applying deconvolution
        operation with specified configurations.
    """
    if isinstance(dimension_size, tf.Dimension):
        dimension_size = dimension_size.value

    if dimension_size is None:
        return None

    if dilation != 1:
        raise ValueError("Deconvolution layer doesn't support dilation")

    if padding in ('VALID', 'valid'):
        return dimension_size * stride + max(filter_size - stride, 0)

    elif padding in ('SAME', 'same'):
        return dimension_size * stride

    elif isinstance(padding, int):
        return dimension_size * stride - 2 * padding + filter_size - 1

    raise ValueError(
        "`{!r}` is unknown deconvolution's padding value".format(padding))


def conv_output_shape(dimension_size, filter_size, padding, stride,
                      dilation=1):
    """
    Computes convolution's output shape for one spatial dimension.

    Parameters
    ----------
    dimension_size : int or None
        Size of the dimension. Typically it's image's weight or height.
        It might be equal to ``None`` when we input might have variable
        dimension.

    filter_size : int
        Size of the convolution filter.

    padding : {``valid``, ``same``} or int
        Type or size of the zero-padding.

    stride : int
        Stride size.

    dilation : int
        Dilation rate. Defaults to ``1``.

    Returns
    -------
    int
        Dimension size after applying convolution
        operation with specified configurations.
    """
    if isinstance(dimension_size, tf.Dimension):
        dimension_size = dimension_size.value

    if dimension_size is None:
        return None

    if not isinstance(stride, int):
        raise ValueError(
            "Stride needs to be an integer, got {} (value {!r})"
            "".format(type(stride), stride))

    if not isinstance(filter_size, int):
        raise ValueError(
            "Filter size needs to be an integer, got {} "
            "(value {!r})".format(type(filter_size), filter_size))

    # We can think of the dilation as very sparse convolutional filter
    # filter=3 and dilation=2 the same as filter=5 and dilation=1
    filter_size = filter_size + (filter_size - 1) * (dilation - 1)

    if padding in ('VALID', 'valid'):
        return int(math.ceil((dimension_size - filter_size + 1) / stride))

    elif padding in ('SAME', 'same'):
        return int(math.ceil(dimension_size / stride))

    elif isinstance(padding, int):
        return int(math.ceil(
            (dimension_size + 2 * padding - filter_size + 1) / stride))

    raise ValueError(
        "`{!r}` is unknown convolution's padding value".format(padding))


class PaddingProperty(Property):
    expected_type = (six.string_types, int, tuple)
    valid_string_choices = ('VALID', 'SAME', 'same', 'valid')

    def __set__(self, instance, value):
        if isinstance(value, int):
            if value < 0:
                raise ValueError(
                    "Integer border mode value needs to be "
                    "greater or equal to zero, got {}".format(value))

            value = (value, value)

        if isinstance(value, six.string_types):
            value = value.upper()

        super(PaddingProperty, self).__set__(instance, value)

    def validate(self, value):
        super(PaddingProperty, self).validate(value)

        if isinstance(value, tuple) and len(value) != 2:
            raise ValueError(
                "Border mode property suppose to get a tuple that "
                "contains two elements, got {} elements"
                "".format(len(value)))

        is_invalid_string = (
            isinstance(value, six.string_types) and
            value not in self.valid_string_choices
        )

        if is_invalid_string:
            valid_choices = ', '.join(self.valid_string_choices)
            raise ValueError(
                "`{}` is invalid string value. Available choices: {}"
                "".format(value, valid_choices))

        if isinstance(value, tuple) and any(element < 0 for element in value):
            raise ValueError(
                "Tuple border mode value needs to contain only elements "
                "that greater or equal to zero, got {}".format(value))


class Convolution(BaseLayer):
    """
    Convolutional layer.

    Parameters
    ----------
    size : tuple of int
        Filter shape. In should be defined as a tuple with three
        integers ``(filter rows, filter columns, output channels)``.

    padding : {{``same``, ``valid``}}, int, tuple
        Zero padding for the input tensor.

        - ``valid`` - Padding won't be added to the tensor. Result will be
          the same as for ``padding=0``

        - ``same`` - Padding will depend on the number of rows and columns
          in the filter. This padding makes sure that image with the
          ``stride=1`` won't change its width and height. It's the same as
          ``padding=(filter rows // 2, filter columns // 2)``.

        - Custom value for the padding can be specified as an integer, like
          ``padding=1`` or it can be specified as a tuple when different
          dimensions have different padding values, for example
          ``padding=(2, 3)``.

        Defaults to ``valid``.

    stride : tuple with ints, int.
        Stride size. Defaults to ``(1, 1)``

    dilation : int, tuple
        Rate for the filter upsampling. When ``dilation > 1`` layer will
        become dilated convolution (or atrous convolution). Defaults to ``1``.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Shape of the weight will be equal to
        ``(filter rows, filter columns, input channels, output channels)``.
        Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`HeNormal(gain=2) <neupy.init.HeNormal>`.

    bias : 1D array-like, Tensorfow variable, scalar, Initializer or None
        Defines layer's bias. Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`Constant(0) <neupy.init.Constant>`.
        The ``None`` value excludes bias from the calculations and
        do not add it into parameters list.

    {BaseLayer.name}

    Examples
    --------
    2D Convolution

    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((28, 28, 3)),
    ...     layers.Convolution((3, 3, 16)),
    ... )

    1D Convolution

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((30, 10)),
    ...     Reshape((30, 1, 10)),  # convert 3D to 4D
    ...     Convolution((3, 1, 16)),
    ...     Reshape((-1, 16))  # convert 4D back to 3D
    ... )
    >>> network
    (?, 30, 10) -> [... 4 layers ...] -> (?, 28, 16)

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    size = TypedListProperty(element_type=int, n_elements=3)
    weight = ParameterProperty()
    bias = ParameterProperty(allow_none=True)

    padding = PaddingProperty()
    stride = Spatial2DProperty()
    dilation = Spatial2DProperty()

    # We use gain=2 because it's suitable choice for relu non-linearity
    # and relu is the most common non-linearity used for CNN.
    def __init__(self, size, padding='valid', stride=1, dilation=1,
                 weight=init.HeNormal(gain=2), bias=0, name=None):

        super(Convolution, self).__init__(name=name)

        self.size = size
        self.padding = padding
        self.stride = stride
        self.dilation = dilation
        self.weight = weight
        self.bias = bias

    def fail_if_shape_invalid(self, input_shape):
        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "Convolutional layer expects an input with 4 "
                "dimensions, got {} with shape {}"
                "".format(len(input_shape), input_shape))

    def output_shape_per_dim(self, *args, **kwargs):
        return conv_output_shape(*args, **kwargs)

    def expected_output_shape(self, input_shape):
        n_samples = input_shape[0]
        row_filter_size, col_filter_size, n_kernels = self.size
        row_stride, col_stride = self.stride
        row_dilation, col_dilation = self.dilation

        if isinstance(self.padding, (list, tuple)):
            row_padding, col_padding = self.padding
        else:
            row_padding, col_padding = self.padding, self.padding

        return (
            n_samples,
            self.output_shape_per_dim(
                input_shape[1], row_filter_size,
                row_padding, row_stride, row_dilation
            ),
            self.output_shape_per_dim(
                input_shape[2], col_filter_size,
                col_padding, col_stride, col_dilation
            ),
            n_kernels,
        )

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        self.fail_if_shape_invalid(input_shape)

        if input_shape.ndims is None:
            n_samples = input_shape[0]
            n_kernels = self.size[-1]
            return tf.TensorShape((n_samples, None, None, n_kernels))

        return tf.TensorShape(self.expected_output_shape(input_shape))

    def create_variables(self, input_shape):
        self.input_shape = input_shape
        n_channels = input_shape[-1]
        n_rows, n_cols, n_filters = self.size

        # Compare to the regular convolution weights,
        # transposed one has switched input and output channels.
        self.weight = self.variable(
            value=self.weight, name='weight',
            shape=(n_rows, n_cols, n_channels, n_filters))

        if self.bias is not None:
            self.bias = self.variable(
                value=self.bias, name='bias',
                shape=as_tuple(n_filters))

    def output(self, input, **kwargs):
        input = tf.convert_to_tensor(input, tf.float32)
        self.fail_if_shape_invalid(input.shape)
        padding = self.padding

        if not isinstance(padding, six.string_types):
            height_pad, width_pad = padding
            input = tf.pad(input, [
                [0, 0],
                [height_pad, height_pad],
                [width_pad, width_pad],
                [0, 0],
            ])
            # VALID option will make sure that
            # convolution won't use any padding.
            padding = 'VALID'

        output = tf.nn.convolution(
            input,
            self.weight,
            padding=padding,
            strides=self.stride,
            dilation_rate=self.dilation,
            data_format="NHWC",
        )

        if self.bias is not None:
            bias = tf.reshape(self.bias, (1, 1, 1, -1))
            output += bias

        return output

    def __repr__(self):
        return self._repr_arguments(
            self.size,
            padding=self.padding,
            stride=self.stride,
            dilation=self.dilation,
            weight=self.weight,
            bias=self.bias,
            name=self.name,
        )


class Deconvolution(Convolution):
    """
    Deconvolution layer (also known as Transposed Convolution.).

    Parameters
    ----------
    {Convolution.size}

    {Convolution.padding}

    {Convolution.stride}

    {Convolution.dilation}

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Shape of the weight will be equal to
        ``(filter rows, filter columns, output channels, input channels)``.
        Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`HeNormal(gain=2) <neupy.init.HeNormal>`.

    {Convolution.bias}

    {Convolution.name}

    Methods
    -------
    {Convolution.Methods}

    Attributes
    ----------
    {Convolution.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 3)),
    ...     Convolution((3, 3, 16)),
    ...     Deconvolution((3, 3, 1)),
    ... )
    >>> network
    (?, 28, 28, 3) -> [... 3 layers ...] -> (?, 28, 28, 1)
    """
    def __init__(self, size, padding='valid', stride=1,
                 weight=init.HeNormal(gain=2), bias=0, name=None):

        super(Deconvolution, self).__init__(
            size=size, padding=padding, stride=stride,
            dilation=1, weight=weight, bias=bias, name=name)

    def output_shape_per_dim(self, *args, **kwargs):
        return deconv_output_shape(*args, **kwargs)

    def create_variables(self, input_shape):
        self.input_shape = input_shape
        n_channels = input_shape[-1]
        n_rows, n_cols, n_filters = self.size

        # Compare to the regular convolution weights,
        # transposed one has switched input and output channels.
        self.weight = self.variable(
            value=self.weight, name='weight',
            shape=(n_rows, n_cols, n_filters, n_channels))

        if self.bias is not None:
            self.bias = self.variable(
                value=self.bias, name='bias',
                shape=as_tuple(n_filters))

    def output(self, input, **kwargs):
        input = tf.convert_to_tensor(input, tf.float32)
        # We need to get information about output shape from the input
        # tensor's shape, because for some inputs we might have
        # height and width specified as None and shape value won't be
        # computed for these dimensions.
        padding = self.padding

        # It's important that expected output shape gets computed on then
        # Tensor (produced by tf.shape) rather than on TensorShape object.
        # Tensorflow cannot convert TensorShape object into Tensor and
        # it will cause an exception in the conv2d_transpose layer.
        output_shape = self.expected_output_shape(tf.shape(input))

        if isinstance(self.padding, (list, tuple)):
            height_pad, width_pad = self.padding

            # VALID option will make sure that
            # deconvolution won't use any padding.
            padding = 'VALID'

            # conv2d_transpose doesn't know about extra paddings that we added
            # in the convolution. For this reason, we have to expand our
            # expected output shape and later we will remove these paddings
            # manually after transpose convolution.
            output_shape = (
                output_shape[0],
                output_shape[1] + 2 * height_pad,
                output_shape[2] + 2 * width_pad,
                output_shape[3],
            )

        output = tf.nn.conv2d_transpose(
            value=input,
            filter=self.weight,
            output_shape=list(output_shape),
            strides=as_tuple(1, self.stride, 1),
            padding=padding,
            data_format="NHWC"
        )

        if isinstance(self.padding, (list, tuple)):
            h_pad, w_pad = self.padding

            if h_pad > 0:
                output = output[:, h_pad:-h_pad, :, :]

            if w_pad > 0:
                output = output[:, :, w_pad:-w_pad, :]

        if self.bias is not None:
            bias = tf.reshape(self.bias, (1, 1, 1, -1))
            output += bias

        return output

    def __repr__(self):
        return self._repr_arguments(
            self.size,
            padding=self.padding,
            stride=self.stride,
            weight=self.weight,
            bias=self.bias,
            name=self.name,
        )



import tensorflow as tf

from neupy import layers
from neupy.utils import tf_utils, as_tuple
from neupy.layers.base import BaseGraph


__all__ = ('mixture_of_experts',)


def check_if_network_is_valid(network, index):
    if not isinstance(network, BaseGraph):
        raise TypeError(
            "Invalid input, Mixture of experts expects networks/layers"
            "in the list of networks, got `{}` instance instead"
            "".format(type(network)))

    if len(network.input_layers) > 1:
        raise ValueError(
            "Each network from the mixture of experts has to process single "
            "input tensor. Network #{} (0-based indices) has more than one "
            "input layer. Input layers: {}"
            "".format(index, network.output_layers))

    if len(network.output_layers) > 1:
        raise ValueError(
            "Each network from the mixture of experts has to output single "
            "tensor. Network #{} (0-based indices) has more than one output "
            "layer. Output layers: {}".format(index, network.output_layers))

    if network.input_shape.ndims != 2:
        raise ValueError(
            "Each network from the mixture of experts has to process "
            "only 2-dimensional inputs. Network #{} (0-based indices) "
            "processes only {}-dimensional inputs. Input layer's shape: {}"
            "".format(index, network.input_shape.ndims, network.input_shape))


def check_if_networks_compatible(networks):
    input_shapes = []
    output_shapes = []

    for i, network in enumerate(networks):
        input_shapes.append(network.input_shape)
        output_shapes.append(network.output_shape)

    for shape in input_shapes:
        if not shape.is_compatible_with(input_shapes[0]):
            raise ValueError(
                "Networks have incompatible input shapes. Shapes: {}"
                "".format(tf_utils.shape_to_tuple(input_shapes)))

    for shape in output_shapes:
        if not shape.is_compatible_with(output_shapes[0]):
            raise ValueError(
                "Networks have incompatible output shapes. Shapes: {}"
                "".format(tf_utils.shape_to_tuple(output_shapes)))


def mixture_of_experts(networks, gating_layer=None):
    """
    Generates mixture of experts architecture from the set of
    networks that has the same input and output shapes.

    Mixture of experts learns to how to mix results from different
    networks in order to get better performances. It adds gating layer
    that using input data tries to figure out which of the networks
    will make better contribution to the final result. The final result
    mixes from all networks using different weights. The higher the weight
    the larger contribution from the individual layer.

    Parameters
    ----------
    networks : list of networks/layers

    gating_layer : None or layer
        In case if value equal to `None` that the following layer
        will be created.

        .. code-block:: python

            gating_layer = layers.Softmax(len(networks))

        Output from the gating layer should be 1D and equal to
        the number of networks.

    Raises
    ------
    ValueError
        In case if there is some problem with input networks
        or custom gating layer.

    Returns
    -------
    network
        Mixture of experts network that combine all networks into
        single one and adds gating layer to it.

    Examples
    --------
    >>> from neupy import algorithms, architectures
    >>> from neupy.layers import *
    >>>
    >>> network = architectures.mixture_of_experts([
    ...     join(
    ...         Input(10),
    ...         Relu(5),
    ...     ),
    ...     join(
    ...         Input(10),
    ...         Relu(33),
    ...         Relu(5),
    ...     ),
    ...     join(
    ...         Input(10),
    ...         Relu(12),
    ...         Relu(25),
    ...         Relu(5),
    ...     ),
    ... ])
    >>> network
    (?, 10) -> [... 12 layers ...] -> (?, 5)
    >>>
    >>> optimizer = algorithms.Momentum(network, step=0.1)
    """
    if not isinstance(networks, (list, tuple)):
        raise ValueError("Networks should be specified as a list")

    for index, network in enumerate(networks):
        check_if_network_is_valid(network, index)

    check_if_networks_compatible(networks)
    input_shape = tf.TensorShape(None)

    for network in networks:
        input_shape = input_shape.merge_with(network.input_shape)

    n_layers_to_combine = len(networks)
    n_features = input_shape[1].value

    if n_features is None:
        raise ValueError(
            "Cannot create mixture of experts model, because "
            "number of input features is unknown")

    if gating_layer is None:
        gating_layer = layers.Softmax(n_layers_to_combine)

    if not isinstance(gating_layer, layers.BaseLayer):
        raise ValueError(
            "Invalid type for gating layer. Type: {}"
            "".format(type(gating_layer)))

    return layers.join(
        layers.Input(n_features),
        # Note: Gating network should be specified
        # as a first parameter.
        layers.parallel(*as_tuple(gating_layer, networks)),
        layers.GatedAverage(),
    )



from __future__ import division

import numpy as np
from sklearn import datasets, metrics
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder

from neupy.layers import *
from neupy import algorithms, layers


X, y = datasets.fetch_openml('mnist_784', version=1, return_X_y=True)
X = X / 255.

target_scaler = OneHotEncoder(categories='auto', sparse=False)
y = target_scaler.fit_transform(y.reshape(-1, 1))

# Originaly we should have 70000 images from the MNIST dataset, but
# we will use only 1000 training example, All data that doesn't have
# labels we use to train features in the convolutional autoencoder.
n_labeled = 1000
n_samples = len(X)
n_unlabeled = n_samples - n_labeled

x_labeled, x_unlabeled, y_labeled, y_unlabeled = train_test_split(
    X.astype(np.float32),
    y.astype(np.float32),
    test_size=(1 - n_labeled / n_samples)
)

x_labeled_4d = x_labeled.reshape((n_labeled, 28, 28, 1))
x_unlabeled_4d = x_unlabeled.reshape((n_unlabeled, 28, 28, 1))

# We will features trained in the encoder and the first part for the future
# classifier. At first we pre-train them with unlabeled data, since we have
# a lot of it and we hope to learn some common features from it.
encoder = join(
    Input((28, 28, 1)),

    Convolution((3, 3, 16)) >> Relu(),
    Convolution((3, 3, 16)) >> Relu(),
    MaxPooling((2, 2)),

    Convolution((3, 3, 32)) >> Relu(),
    MaxPooling((2, 2)),

    Reshape(),

    Relu(256),
    Relu(128),
)

# Notice that in the decoder every operation reverts
# back changes from the encoder layer.
decoder = join(
    Relu(256),
    Relu(32 * 5 * 5),

    Reshape((5, 5, 32)),

    Upscale((2, 2)),
    Deconvolution((3, 3, 16)) >> Relu(),

    Upscale((2, 2)),
    Deconvolution((3, 3, 16)) >> Relu(),
    Deconvolution((3, 3, 1)) >> Sigmoid(),
)

conv_autoencoder = algorithms.Momentum(
    network=(encoder >> decoder),

    loss='rmse',
    step=0.02,
    batch_size=128,
    regularizer=algorithms.l2(0.001),

    shuffle_data=True,
    verbose=True,
)
conv_autoencoder.train(
    x_unlabeled_4d, x_unlabeled_4d,
    x_labeled_4d, x_labeled_4d,
    epochs=1,
)

# In order to speed up training for the upper layers we generate
# output from the encoder. In this way we won't need to regenerate
# encoded inputs for every epoch.
x_labeled_encoded = encoder.predict(x_labeled_4d)
x_unlabeled_encoded = encoder.predict(x_unlabeled_4d)

classifier_network = join(
    PRelu(512),
    Dropout(0.25),
    Softmax(10),
)

encoder_classifier = algorithms.Adadelta(
    Input(encoder.output_shape[1:]) >> classifier_network,
    verbose=True,
    step=0.05,
    shuffle_data=True,
    batch_size=64,
    loss='categorical_crossentropy',
)
encoder_classifier.train(
    x_labeled_encoded, y_labeled,
    x_unlabeled_encoded, y_unlabeled,
    epochs=400,
)

# The final part of training is to put encoder and final classifier layers
# in order to fine tune network parameters before finilizing it's prediction
classifier = algorithms.GradientDescent(
    network=(encoder >> classifier_network),
    verbose=True,
    step=0.005,
    shuffle_data=True,
    batch_size=64,
    loss='categorical_crossentropy',
    regularizer=algorithms.l2(0.02),
)
classifier.train(x_labeled_4d, y_labeled, epochs=100)
classifier.train(
    x_labeled_4d, y_labeled,
    x_unlabeled_4d, y_unlabeled,
    epochs=1,
)

unlabeled_predicted = classifier.predict(x_unlabeled_4d).argmax(axis=1)
y_unlabeled_classes = np.asarray(y_unlabeled).argmax(axis=1)

print(metrics.classification_report(y_unlabeled_classes, unlabeled_predicted))
score = metrics.accuracy_score(y_unlabeled_classes, unlabeled_predicted)
print("Validation accuracy: {:.2%}".format(score))



from __future__ import division

import time

import numpy as np
import tensorflow as tf

from neupy import layers
from neupy.core.properties import (
    FunctionWithOptionsProperty,
    ScalarVariableProperty,
    IntProperty, Property,
)
from neupy.utils import (
    AttributeKeyDict, format_data,
    as_tuple, iters, tf_utils,
)
from neupy.algorithms.gd import objectives
from neupy.exceptions import InvalidConnection
from neupy.algorithms.base import BaseNetwork


__all__ = ('BaseOptimizer', 'GradientDescent')


class BaseOptimizer(BaseNetwork):
    """
    Gradient descent algorithm.

    Parameters
    ----------
    network : list, tuple or LayerConnection instance
        Network's architecture. There are a few ways
        to define it.

        - List of layers.
          For instance, ``[Input(2), Tanh(4), Relu(1)]``.

        - Constructed layers.
          For instance, ``Input(2) >> Tanh(4) >> Relu(1)``.

    regularizer : function or None
        Network's regularizer.

    loss : str or function
        Error/loss function. Defaults to ``mse``.

        - ``mae`` - Mean Absolute Error.

        - ``mse`` - Mean Squared Error.

        - ``rmse`` - Root Mean Squared Error.

        - ``msle`` - Mean Squared Logarithmic Error.

        - ``rmsle`` - Root Mean Squared Logarithmic Error.

        - ``categorical_crossentropy`` - Categorical cross entropy.

        - ``binary_crossentropy`` - Binary cross entropy.

        - ``binary_hinge`` - Binary hinge entropy.

        - ``categorical_hinge`` - Categorical hinge entropy.

        - Custom function which accepts two mandatory arguments.
          The first one is expected value and the second one is
          predicted value. Example:

        .. code-block:: python

            def custom_func(expected, predicted):
                return expected - predicted

    step : float, Variable
        Learning rate, defaults to ``0.1``.

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {BaseNetwork.verbose}

    Attributes
    ----------
    {BaseNetwork.Attributes}

    Methods
    -------
    {BaseSkeleton.predict}

    train(X_train, y_train, X_test=None, y_test=None, epochs=100)
        Train network. You can control network's training procedure
        with ``epochs`` parameter. The ``X_test`` and ``y_test`` should
        be presented both in case network's validation required
        after each training epoch.

    {BaseSkeleton.fit}
    """
    step = ScalarVariableProperty(default=0.1)
    target = Property(default=None, allow_none=True)
    regularizer = Property(default=None, allow_none=True)
    loss = FunctionWithOptionsProperty(default='mse', choices={
        'mae': objectives.mae,
        'mse': objectives.mse,
        'rmse': objectives.rmse,
        'msle': objectives.msle,
        'rmsle': objectives.rmsle,

        'binary_crossentropy': objectives.binary_crossentropy,
        'categorical_crossentropy': objectives.categorical_crossentropy,

        'binary_hinge': objectives.binary_hinge,
        'categorical_hinge': objectives.categorical_hinge,
    })

    def __init__(self, network, options=None, **kwargs):
        options = options or kwargs

        if isinstance(network, (list, tuple)):
            network = layers.join(*network)

        self.network = network

        if len(self.network.output_layers) != 1:
            n_outputs = len(network.output_layers)

            raise InvalidConnection(
                "Connection should have one output "
                "layer, got {}".format(n_outputs))

        target = options.get('target')
        if target is not None and isinstance(target, (list, tuple)):
            options['target'] = tf.placeholder(tf.float32, shape=target)

        self.target = self.network.targets
        super(BaseOptimizer, self).__init__(**options)

        start_init_time = time.time()
        self.logs.message(
            "TENSORFLOW",
            "Initializing Tensorflow variables and functions.")

        self.variables = AttributeKeyDict()
        self.functions = AttributeKeyDict()
        self.network.outputs
        self.init_functions()

        self.logs.message(
            "TENSORFLOW",
            "Initialization finished successfully. It took {:.2f} seconds"
            "".format(time.time() - start_init_time))

    def init_train_updates(self):
        raise NotImplementedError()

    def init_functions(self):
        loss = self.loss(self.target, self.network.outputs)
        val_loss = self.loss(self.target, self.network.training_outputs)

        if self.regularizer is not None:
            loss += self.regularizer(self.network)

        self.variables.update(
            step=self.step,
            loss=loss,
            val_loss=val_loss,
        )

        with tf.name_scope('training-updates'):
            update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)

            with tf.control_dependencies(update_ops):
                training_updates = self.init_train_updates()
                training_updates.extend(update_ops)

        tf_utils.initialize_uninitialized_variables()

        with tf.name_scope('optimizer'):
            self.functions.update(
                predict=tf_utils.function(
                    inputs=as_tuple(self.network.inputs),
                    outputs=self.network.outputs,
                    name='predict'
                ),
                one_training_update=tf_utils.function(
                    inputs=as_tuple(self.network.inputs, self.target),
                    outputs=loss,
                    updates=training_updates,
                    name='one-update-step'
                ),
                score=tf_utils.function(
                    inputs=as_tuple(self.network.inputs, self.target),
                    outputs=val_loss,
                    name='score'
                ),
            )

    def format_input(self, X):
        X = as_tuple(X)
        X_formatted = []

        if len(X) != len(self.network.input_layers):
            raise ValueError(
                "Number of inputs doesn't match number "
                "of input layers in the network.")

        for input, input_layer in zip(X, self.network.input_layers):
            input_shape = tf.TensorShape(input_layer.input_shape)
            is_feature1d = (input_shape.ndims == 2 and input_shape[1] == 1)
            formatted_input = format_data(input, is_feature1d=is_feature1d)

            if (formatted_input.ndim + 1) == input_shape.ndims:
                # We assume that when one dimension was missed than user
                # wants to propagate single sample through the network
                formatted_input = np.expand_dims(formatted_input, axis=0)

            X_formatted.append(formatted_input)

        return X_formatted

    def format_target(self, y):
        output_shape = tf.TensorShape(self.network.output_shape)
        is_feature1d = (output_shape.ndims == 2 and output_shape[1] == 1)
        formatted_target = format_data(y, is_feature1d=is_feature1d)

        if (formatted_target.ndim + 1) == len(output_shape):
            # We assume that when one dimension was missed than user
            # wants to propagate single sample through the network
            formatted_target = np.expand_dims(formatted_target, axis=0)

        return formatted_target

    def score(self, X, y):
        """
        Calculate prediction accuracy for input data.

        Parameters
        ----------
        X : array-like
        y : array-like

        Returns
        -------
        float
            Prediction error.
        """
        X = self.format_input(X)
        y = self.format_target(y)
        return self.functions.score(*as_tuple(X, y))

    def predict(self, *X, **kwargs):
        """
        Makes a raw prediction.

        Parameters
        ----------
        X : array-like

        Returns
        -------
        array-like
        """
        default_batch_size = getattr(self, 'batch_size', None)
        predict_kwargs = dict(
            batch_size=kwargs.pop('batch_size', default_batch_size),
            verbose=self.verbose,
        )

        # We require do to this check for python 2 compatibility
        if kwargs:
            raise TypeError("Unknown arguments: {}".format(kwargs))

        return self.network.predict(*self.format_input(X), **predict_kwargs)

    def train(self, X_train, y_train, X_test=None, y_test=None,
              *args, **kwargs):

        is_test_data_partialy_missing = (
            (X_test is None and y_test is not None) or
            (X_test is not None and y_test is None)
        )

        if is_test_data_partialy_missing:
            raise ValueError(
                "Input or target test samples are missed. They "
                "must be defined together or none of them.")

        X_train = self.format_input(X_train)
        y_train = self.format_target(y_train)

        if X_test is not None:
            X_test = self.format_input(X_test)
            y_test = self.format_target(y_test)

        return super(BaseOptimizer, self).train(
            X_train=X_train, y_train=y_train,
            X_test=X_test, y_test=y_test,
            *args, **kwargs)

    def one_training_update(self, X_train, y_train):
        return self.functions.one_training_update(
            *as_tuple(X_train, y_train))

    def get_params(self, deep=False, with_network=True):
        params = super(BaseOptimizer, self).get_params()
        if with_network:
            params['network'] = self.network
        return params

    def __reduce__(self):
        parameters = self.get_params(with_network=False)

        # We only need to know placeholders shape
        # in order to be able to reconstruct it
        parameters['target'] = tf_utils.shape_to_tuple(
            parameters['target'].shape)

        args = (self.network, parameters)
        return (self.__class__, args)

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__,
            self.network,
            self.repr_options())


class GradientDescent(BaseOptimizer):
    """
    Mini-batch Gradient Descent algorithm.

    Parameters
    ----------
    batch_size : int or None
        Set up min-batch size. The ``None`` value will ensure that all data
        samples will be propagated through the network at once.
        Defaults to ``128``.

    {BaseOptimizer.Parameters}

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.algorithms import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.GradientDescent(network, batch_size=1)
    >>> optimizer.train(x_train, y_train)
    """
    batch_size = IntProperty(default=128, minval=0, allow_none=True)

    def init_train_updates(self):
        optimizer = tf.train.GradientDescentOptimizer(
            learning_rate=self.step,
        )
        self.functions.optimizer = optimizer
        return [optimizer.minimize(self.variables.loss)]

    def one_training_update(self, X_train, y_train):
        """
        Train one epoch.

        Parameters
        ----------
        X_train : array-like
            Training input dataset.

        y_train : array-like
            Training target dataset.

        Returns
        -------
        float
            Training error.
        """
        return self.functions.one_training_update(
            *as_tuple(X_train, y_train))

    def score(self, X, y):
        """
        Check the prediction error for the specified input samples
        and their targets.

        Parameters
        ----------
        X : array-like
        y : array-like

        Returns
        -------
        float
            Prediction error.
        """
        X = self.format_input(X)
        y = self.format_target(y)

        return iters.apply_batches(
            function=self.functions.score,
            inputs=as_tuple(X, y),
            batch_size=self.batch_size,
            show_output=True,
            show_progressbar=self.logs.enable,
            average_outputs=True,
        )



import pkgutil
import warnings
import importlib

import matplotlib.pyplot as plt


def load_pandas_module():
    if not pkgutil.find_loader('pandas'):
        raise ImportError(
            "The `pandas` library is not installed. Try to "
            "install it with pip: \n    pip install pandas")

    return importlib.import_module('pandas')


def plot(x_train, y_train, x_valid, y_valid, ax, logx=False):
    plot_function = ax.semilogx if logx else ax.plot

    # With large number of samples there will be to many
    # points in the plot and it will be hard to read it
    style = '-' if len(x_train) > 50 or len(x_valid) > 50 else 'o-'
    line_train, = plot_function(x_train, y_train, style, label='train')

    if len(x_valid) > 0:
        line_valid_, = plot_function(x_valid, y_valid, style, label='valid')
        ax.legend([line_train, line_valid_], ['Training', 'Validation'])


def plot_errors_per_update(train, valid, ax, logx=False):
    plot(
        # Loss that we've obtain per each training update is calculated before
        # the update. Which means that loss represents state of the network
        # before the update.
        x_train=train.n_updates - 1,
        y_train=train.value,

        x_valid=valid.n_updates,
        y_valid=valid.value,

        logx=logx,
        ax=ax,
    )

    ax.set_title('Training perfomance')
    ax.set_ylabel('Loss')
    ax.set_xlabel('Number of updates passed')


def plot_error_per_epoch(train, valid, ax, logx=False):
    valid = valid[['epoch', 'value']]
    train = train[['epoch', 'value']]

    train_epochs = train.groupby('epoch').agg({'value': 'first'})
    valid_epochs = valid.groupby('epoch').agg({'value': 'first'})

    plot(
        # Loss that we've obtain per each training update is calculated before
        # the update. Which means that loss represents state of the network
        # before the update.
        x_train=train_epochs.index - 1,
        y_train=train_epochs.value,

        x_valid=valid_epochs.index,
        y_valid=valid_epochs.value,

        logx=logx,
        ax=ax,
    )

    ax.set_ylabel('Loss')
    ax.set_xlabel('Number of training epochs passed')


def plot_optimizer_errors(optimizer, logx=False, show=True, **figkwargs):
    if 'figsize' not in figkwargs:
        figkwargs['figsize'] = (12, 8)

    if not optimizer.events.logs:
        warnings.warn("There is no data to plot")
        return

    pd = load_pandas_module()
    history = pd.DataFrame(optimizer.events.logs)

    train = history[history.name == 'train_error']
    valid = history[history.name == 'valid_error']
    train_max = train.max()

    if train_max.epoch == train_max.n_updates:
        fig, ax = plt.subplots(1, 1, **figkwargs)
        # When number of epochs exactly the same as number of updates
        # plots per number of updates and epochs will be exactly the same
        plot_error_per_epoch(train, valid, logx=logx, ax=ax)
    else:
        fig, axes = plt.subplots(2, 1, **figkwargs)
        plot_errors_per_update(train, valid, logx=logx, ax=axes[0])
        plot_error_per_epoch(train, valid, logx=logx, ax=axes[1])

    if show:
        plt.show()



from __future__ import division

import math

import tensorflow as tf

from neupy.utils import as_tuple, tf_utils
from neupy.core.properties import (TypedListProperty, ChoiceProperty,
                                   FunctionWithOptionsProperty)
from neupy.exceptions import LayerConnectionError
from .base import BaseLayer
from .convolutions import Spatial2DProperty


__all__ = ('MaxPooling', 'AveragePooling', 'Upscale', 'GlobalPooling')


def pooling_output_shape(dimension_size, pool_size, padding, stride):
    """
    Computes output shape for pooling operation.

    Parameters
    ----------
    dimension_size : int
        Size of the dimension. Typically it's image's
        weight or height.

    filter_size : int
        Size of the pooling filter.

    padding : int
        Size of the zero-padding.

    stride : int
        Stride size.

    Returns
    -------
    int
    """
    if isinstance(dimension_size, tf.Dimension):
        dimension_size = dimension_size.value

    if dimension_size is None:
        return None

    if padding in ('SAME', 'same'):
        return int(math.ceil(dimension_size / stride))

    elif padding in ('VALID', 'valid'):
        return int(math.ceil((dimension_size - pool_size + 1) / stride))

    raise ValueError(
        "{!r} is unknown padding value for pooling".format(padding))


class BasePooling(BaseLayer):
    """
    Base class for the pooling layers.

    Parameters
    ----------
    size : tuple with 2 integers
        Factor by which to downscale ``(vertical, horizontal)``.
        ``(2, 2)`` will halve the image in each dimension.

    stride : tuple or int.
        Stride size, which is the number of shifts over
        rows/cols to get the next pool region. If stride is
        ``None``, it is considered equal to ``size`` (no overlap on
        pooling regions).

    padding : {{``valid``, ``same``}}
        ``(pad_h, pad_w)``, pad zeros to extend beyond four borders of
        the images, pad_h is the size of the top and bottom margins,
        and pad_w is the size of the left and right margins.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    size = TypedListProperty(required=True, element_type=int)
    stride = Spatial2DProperty(allow_none=True)
    padding = ChoiceProperty(choices=('SAME', 'VALID', 'same', 'valid'))
    pooling_type = None

    def __init__(self, size, stride=None, padding='valid', name=None):
        super(BasePooling, self).__init__(name=name)

        self.size = size
        self.stride = stride
        self.padding = padding

    def fail_if_shape_invalid(self, input_shape):
        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "Pooling layer expects an input with 4 "
                "dimensions, got {} with shape {}. Layer: {}"
                "".format(len(input_shape), input_shape, self))

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if input_shape.ndims is None:
            return tf.TensorShape((None, None, None, None))

        self.fail_if_shape_invalid(input_shape)

        n_samples, rows, cols, n_kernels = input_shape
        row_filter_size, col_filter_size = self.size

        stride = self.size if self.stride is None else self.stride
        row_stride, col_stride = stride

        output_rows = pooling_output_shape(
            rows, row_filter_size, self.padding, row_stride)

        output_cols = pooling_output_shape(
            cols, col_filter_size, self.padding, col_stride)

        # In python 2, we can get float number after rounding procedure
        # and it might break processing in the subsequent layers.
        return tf.TensorShape((n_samples, output_rows, output_cols, n_kernels))

    def output(self, input_value, **kwargs):
        return tf.nn.pool(
            input_value,
            self.size,
            pooling_type=self.pooling_type,
            padding=self.padding.upper(),
            strides=self.stride or self.size,
            data_format="NHWC")

    def __repr__(self):
        return self._repr_arguments(
            self.size,
            name=self.name,
            stride=self.stride,
            padding=self.padding,
        )


class MaxPooling(BasePooling):
    """
    Maximum pooling layer.

    Parameters
    ----------
    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((10, 10, 3)),
    ...     MaxPooling((2, 2)),
    ... )
    >>> network
    (?, 10, 10, 3) -> [... 2 layers ...] -> (?, 5, 5, 3)

    1D pooling

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((30, 10)),
    ...     Reshape((30, 1, 10)),
    ...     MaxPooling((2, 1)),
    ...     Reshape((-1, 10))
    ... )
    >>> network
    (?, 30, 10) -> [... 4 layers ...] -> (?, 15, 10)
    """
    pooling_type = 'MAX'


class AveragePooling(BasePooling):
    """
    Average pooling layer.

    Parameters
    ----------
    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((10, 10, 3)),
    ...     AveragePooling((2, 2)),
    ... )
    >>> network
    (?, 10, 10, 3) -> [... 2 layers ...] -> (?, 5, 5, 3)

    1D pooling

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((30, 10)),
    ...     Reshape((30, 1, 10)),
    ...     AveragePooling((2, 1)),
    ...     Reshape((-1, 10))
    ... )
    >>> network
    (?, 30, 10) -> [... 4 layers ...] -> (?, 15, 10)
    """
    pooling_type = 'AVG'


class Upscale(BaseLayer):
    """
    Upscales input over two axis (height and width).

    Parameters
    ----------
    scale : int or tuple with two int
        Scaling factor for the input value. In the tuple first
        parameter identifies scale of the height and the second
        one of the width.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((10, 10, 3)) >> Upscale((2, 2))
    (?, 10, 10, 3) -> [... 2 layers ...] -> (?, 20, 20, 3)
    """
    scale = TypedListProperty(n_elements=2)

    def __init__(self, scale, name=None):
        super(Upscale, self).__init__(name=name)

        if isinstance(scale, int):
            scale = as_tuple(scale, scale)

        if any(element <= 0 for element in scale):
            raise ValueError(
                "Only positive integers are allowed for scale")

        self.scale = scale

    def fail_if_shape_invalid(self, input_shape):
        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "Upscale layer should have an input value with 4 dimensions "
                "(batch, height, width, channel), got input with {} "
                "dimensions instead. Shape: {}"
                "".format(input_shape.ndims, input_shape))

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        self.fail_if_shape_invalid(input_shape)

        if input_shape.ndims is None:
            return tf.TensorShape((None, None, None, None))

        n_samples, height, width, channel = input_shape
        height_scale, width_scale = self.scale

        return tf.TensorShape([
            n_samples,
            height_scale * height,
            width_scale * width,
            channel,
        ])

    def output(self, input_value, **kwargs):
        input_value = tf.convert_to_tensor(input_value, dtype=tf.float32)
        self.fail_if_shape_invalid(input_value.shape)
        return tf_utils.repeat(input_value, as_tuple(1, self.scale, 1))

    def __repr__(self):
        return self._repr_arguments(self.scale, name=self.name)


class GlobalPooling(BaseLayer):
    """
    Global pooling layer.

    Parameters
    ----------
    function : {{``avg``, ``max``, ``sum``}} or callable
        Common functions has been predefined for the user.
        These options are available:

        - ``avg`` - For average global pooling. The same as
          ``tf.reduce_mean``.

        - ``max`` - For max global pooling. The same as
          ``tf.reduce_max``.

        - ``sum`` - For sum global pooling. The same as
          ``tf.reduce_sum``.

        Parameter also excepts custom functions that have
        following format.

        .. code-block:: python

            def agg_func(x, axis=None):
                pass

        Defaults to ``avg``.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((4, 4, 16)) >> GlobalPooling('avg')
    (?, 4, 4, 16) -> [... 2 layers ...] -> (?, 16)
    """
    function = FunctionWithOptionsProperty(choices={
        'avg': tf.reduce_mean,
        'max': tf.reduce_max,
        'sum': tf.reduce_sum,
    })

    def __init__(self, function, name=None):
        super(GlobalPooling, self).__init__(name=name)
        self.original_function = function
        self.function = function

    def get_output_shape(self, input_shape):
        return tf.TensorShape([input_shape[0], input_shape[-1]])

    def output(self, input_value, **kwargs):
        input_value = tf.convert_to_tensor(input_value, dtype=tf.float32)
        ndims = len(input_value.shape)

        if ndims == 2:
            return input_value

        # All dimensions except first and last
        agg_axis = range(1, ndims - 1)
        return self.function(input_value, axis=list(agg_axis))

    def __repr__(self):
        return self._repr_arguments(
            repr(self.original_function), name=self.name)



import numpy as np
import matplotlib.pyplot as plt
from sklearn import datasets
from neupy import algorithms, layers, utils


def load_data():
    X, _ = datasets.fetch_openml('mnist_784', version=1, return_X_y=True)
    X = (X / 255.).astype(np.float32)

    np.random.shuffle(X)
    x_train, x_test = X[:60000], X[60000:]

    return x_train, x_test


def visualize_reconstructions(autoencoder, x_test):
    n_samples = 4
    image_vectors = x_test[:n_samples, :]
    images = image_vectors * 255.
    predicted_images = autoencoder.predict(image_vectors)
    predicted_images = predicted_images * 255.

    # Compare real and reconstructed images
    fig, axes = plt.subplots(4, 2, figsize=(12, 8))
    iterator = zip(axes, images, predicted_images)

    for (left_ax, right_ax), real_image, predicted_image in iterator:
        real_image = real_image.reshape((28, 28))
        predicted_image = predicted_image.reshape((28, 28))

        left_ax.imshow(real_image, cmap=plt.cm.binary)
        right_ax.imshow(predicted_image, cmap=plt.cm.binary)

    plt.show()


if __name__ == '__main__':
    autoencoder = algorithms.Momentum(
        [
            layers.Input(784),
            layers.GaussianNoise(mean=0.5, std=0.1),
            layers.Sigmoid(100),
            layers.Sigmoid(784),
        ],
        step=0.1,
        verbose=True,
        momentum=0.9,
        nesterov=True,
        loss='rmse',
    )

    print("Preparing data...")
    x_train, x_test = load_data()

    print("Training autoencoder...")
    autoencoder.train(x_train, x_train, x_test, x_test, epochs=40)

    visualize_reconstructions(autoencoder, x_test)



from neupy import layers
from neupy.utils import function_name_scope


__all__ = ('squeezenet',)


@function_name_scope
def Fire(s_1x1, e_1x1, e_3x3, name):
    return layers.join(
        layers.Convolution(
            (1, 1, s_1x1),
            padding='SAME',
            name=name + '/squeeze1x1'
        ),
        layers.Relu(),
        layers.parallel([
            layers.Convolution(
                (1, 1, e_1x1),
                padding='SAME',
                name=name + '/expand1x1'
            ),
            layers.Relu(),
        ], [
            layers.Convolution(
                (3, 3, e_3x3),
                padding='SAME',
                name=name + '/expand3x3'
            ),
            layers.Relu(),
        ]),
        layers.Concatenate(),
    )


def squeezenet():
    """
    SqueezeNet network architecture with random parameters.
    Parameters can be loaded using ``neupy.storage`` module.

    SqueezeNet has roughly 1.2 million parameters. It is almost
    50 times less than in AlexNet. Parameters can be stored as 5Mb
    file.

    Examples
    --------
    >>> from neupy import architectures
    >>> squeezenet = architectures.squeezenet()
    >>> squeezenet
    (?, 227, 227, 3) -> [... 67 layers ...] -> (?, 1000)
    >>>
    >>> from neupy import algorithms
    >>> optimizer = algorithms.Momentum(squeezenet)

    See Also
    --------
    :architecture:`vgg16` : VGG16 network
    :architecture:`vgg19` : VGG19 network
    :architecture:`resnet50` : ResNet50 network

    References
    ----------
    SqueezeNet: AlexNet-level accuracy with 50x fewer parameters
    and <0.5MB model size
    https://arxiv.org/abs/1602.07360
    """
    return layers.join(
        layers.Input((227, 227, 3)),

        layers.Convolution((7, 7, 96), stride=(2, 2),
                           padding='VALID', name='conv1'),
        layers.Relu(),
        layers.MaxPooling((3, 3), stride=(2, 2)),

        Fire(16, 64, 64, name='fire2'),
        Fire(16, 64, 64, name='fire3'),
        Fire(32, 128, 128, name='fire4'),
        layers.MaxPooling((2, 2)),

        Fire(32, 128, 128, name='fire5'),
        Fire(48, 192, 192, name='fire6'),
        Fire(48, 192, 192, name='fire7'),
        Fire(64, 256, 256, name='fire8'),
        layers.MaxPooling((2, 2)),

        Fire(64, 256, 256, name='fire9'),
        layers.Dropout(0.5),

        layers.Convolution((1, 1, 1000), name='conv10'),
        layers.GlobalPooling('avg'),
        layers.Reshape(),
        layers.Softmax(),
    )
